# -*- coding: utf-8 -*-
"""
Description
------------

Functions to deal with numpy arrays:

- :py:func:`~radis.misc.arrays.norm`
- :py:func:`~radis.misc.arrays.norm_on`
- :py:func:`~radis.misc.arrays.scale_to`
- :py:func:`~radis.misc.arrays.array_allclose`
- :py:func:`~radis.misc.arrays.nantrapz`
- :py:func:`~radis.misc.arrays.arange_len`
- :py:func:`~radis.misc.arrays.calc_diff`
- :py:func:`~radis.misc.arrays.find_nearest`
- :py:func:`~radis.misc.arrays.find_first`
- :py:func:`~radis.misc.arrays.autoturn`
- :py:func:`~radis.misc.arrays.centered_diff`
- :py:func:`~radis.misc.arrays.evenly_distributed`
- :py:func:`~radis.misc.arrays.anynan`
- :py:func:`~radis.misc.arrays.first_nonnan_index`
- :py:func:`~radis.misc.arrays.last_nonan_index`
- :py:func:`~radis.misc.arrays.is_sorted`
- :py:func:`~radis.misc.arrays.is_sorted_backward`
- :py:func:`~radis.misc.arrays.bining`
- :py:func:`~radis.misc.arrays.count_nans`
- :py:func:`~radis.misc.arrays.logspace`
- :py:func:`~radis.misc.arrays.numpy_add_at`





-------------------------------------------------------------------------------

"""


from math import ceil

import numba
import numpy as np
from numpy import hstack
from scipy.interpolate import interp1d

# Normalize


def norm(a, normby=None, how="max"):
    """Normalize a numpy array with its maximum. Or normalize it with another
    vector. Works if array contains nans.

    Parameters
    ----------

    normby: array, or None
        if array, norm with this other array's maximum. If None, normalize with
        its maximum.
    """
    if normby is not None:
        normby = np.abs(normby)
    else:
        normby = np.abs(a)

    if how == "max":
        return a / np.nanmax(normby)
    elif how == "mean":
        return a / np.nanmean(normby)
    else:
        raise ValueError("Unknown normalisation method")


def norm_on(a, w, wmin=None, wmax=None, how="max"):
    """Normalize `a` on a specific range of `w`

    Parameters
    ----------

    a: array
        array
    w: array
        x-axis array

    Other Parameters
    ----------------

    wmin, wmax: float
        crop range
    how: 'mean', 'max'
        how to normalize

    Returns
    -------

    a_norm: array
        normalized array
    """
    imin = np.argmin((w - wmin) ** 2) if wmin else None
    imax = np.argmin((w - wmax) ** 2) if wmax else None
    if imin is not None and imax is not None:
        if imin > imax:
            imin, imax = imax, imin
        elif imin == imax:
            imax += 1
    return norm(a, a[imin:imax], how=how)


def scale_to(a, b, k=1):
    """Scale function a to k*b."""
    return a * k * max(np.abs(b)) / max(np.abs(a))


def array_allclose(a, b, rtol=1e-5, atol=1e-8, equal_nan=True):
    """Returns wheter a and b are all close (element wise). If not the same
    size, returns False (instead of crashing like the numpy version). Cf
    numpy.allclose docs for more information.

    Parameters
    ----------

    a, b: arrays
    rtol: float
    atol: float
    equal_nan: bool
        whether to consider Nan's as equal. Contrary to the numpy version this
        one is set to True by default
    """

    if len(a) != len(b):
        return False

    return np.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)


def nantrapz(I, w, dx=1.0, axis=-1):
    """Returns :py:func:`~numpy.trapz` (I, w) discarding nan."""
    b = ~np.isnan(I)
    return np.trapz(I[b], w[b], dx=dx, axis=axis)


# %%
# ==============================================================================
# Numpy Function
# ==============================================================================


def arange_len(wmin, wmax, wstep) -> int:
    """Returns len of a :py:func:`numpy.arange` ``(wmin, max, wstep)`` array,
    accounting for floating point errors

    Note: :py:func:`numpy.arange` is useful to maintain the input ``wstep``.
    If you don't have this requirement, you better use :py:func:`numpy.linspace`
    directly.
    """
    return ceil((wmax - wmin) / wstep)


def calc_diff(t1, v1, t2, v2):
    """Substract two vectors that may have slightly offset abscisses
    interpolating the correct values.

    Parameters
    ----------

    t1, v1: array_like
        first vector and its abscisses
    t2, v2: array_like
        second vector and its abscisses


    Returns
    -------

    tdiff, vdiff: array_like
        substracted vector and its abscisses
    """

    t1, v1, t2, v2 = list(map(np.array, (t1, v1, t2, v2)))

    # Deal with inversely sorted case
    if t1[-1] < t1[0]:
        t1, v1 = t1[::-1], v1[::-1]
    if t2[-1] < t2[0]:
        t2, v2 = t2[::-1], v2[::-1]

    # Get the overlapping range
    b = np.logical_and(t2 > t1[0], t2 < t1[-1])

    tdiff = t2[b]
    v2 = v2[b]

    # Interpolate the correct values
    f = interp1d(t1, v1)
    v1 = f(tdiff)

    # Finally, substract:
    vdiff = v1 - v2

    return tdiff, vdiff


def find_nearest(array, searched, return_bool=False):
    """Return the closest elements in array for each element in 'searched'
    array. In case of multiple elements in `array` having equal difference with
    `searched` element, one with least index is returned. Also returns a
    boolean array with indices of elements occuring in output list set to true.

    Examples
    --------

    ::

        from numpy import array
        find_nearest(array([1,2,3,4]), array([2.1,2]), True)

        >>> (array([2, 2]), array([False, True, False, False]))

        find_nearest(np.array([1,2,3,4]), np.array([2.6,2]), True)

        >>> (array([3, 2]), array([False,  True,  True, False]))

        find_nearest(np.array([1, 3]), np.array([2]))

        >>> array([1])

        find_nearest(np.array([3, 1]), np.array([2]))

        >>> array([3])
    """
    if len(array) == 0:
        raise ValueError("Array to be searched cannot be empty")

    b = np.zeros_like(array, dtype=bool)

    def find_nearest(array, value):
        idx = (np.abs(array - value)).argmin()
        return idx, array[idx]

    nearest_els = []
    try:
        for s in searched:
            idx, el = find_nearest(array, s)
            b[idx] = True
            nearest_els.append(el)
    except:
        idx, el = find_nearest(array, searched)
        b[idx] = True
        nearest_els.append(el)

    if return_bool:
        out = nearest_els, b
    else:
        out = nearest_els

    return out


def find_first(arr, treshold):
    """Return the index of the first element of the array arr whose value is
    more than the treshold."""

    return np.argmax(arr > treshold)


def autoturn(data, key=-1):
    """Turns array data. key value:

    - ``0`` don't transpose
    - ``1`` : transpose
    - ``-1`` : auto : make sure the vectors are along the longest dimension
    """

    if key == 0:
        return data
    elif key == 1:
        return data.transpose()
    elif key == -1:
        if data.shape[0] == max(data.shape):
            # Probably columns :
            return data.transpose()
        else:
            return data


def centered_diff(w):
    """Return w[i+1]-w[i-1]/2, same size as w.

    Similar to :py:func:`numpy.diff`, but does not change the array
    size.
    """
    dw = np.diff(w)
    return (hstack((dw, dw[-1])) + hstack((dw[0], dw))) / 2


def evenly_distributed(w, tolerance=1e-5):
    """Make sure array `w` is evenly distributed.

    Parameters
    ----------

    w : numpy array
        array to test
    tolerance: float
        absolute tolerance

    Returns
    -------

    out: bool
        ``True`` or ``False`` if ``w`` is evenly distributed.
    """
    mean_step = np.diff(w).mean()
    return (np.abs((np.diff(w) - mean_step)) > tolerance).sum() == 0


def anynan(a):
    """Returns whether ``a`` has at least one :py:attr:`~numpy.nan`

    Fastest implementation for arrays with >10^4 elements
    https://stackoverflow.com/a/45011547/5622825
    """
    return np.isnan(np.dot(a, a))


@numba.njit
def first_nonnan_index(a):
    """Returns index of first non-nan value in ``a``

    Returns None is all values are :py:attr:`~numpy.nan`

    See Also
    --------

    :func:`~radis.misc.arrays.last_nonnan_index`
    """
    for i in range(a.size):
        if not np.isnan(a[i]):
            return i
    return None


@numba.njit
def last_nonnan_index(a):
    """Returns index of first non-nan value in ``a``

    Returns None is all values are :py:attr:`~numpy.nan`

    See Also
    --------

    :func:`~radis.misc.arrays.first_nonnan_index`
    """
    for i in range(a.size - 1, 0, -1):
        if not np.isnan(a[i]):
            return i
    return None


@numba.njit
def is_sorted(a):
    """Returns whether ``a`` is sorted in ascending order.

    From B.M. answer on StackOverflow: https://stackoverflow.com/a/47004533/5622825

    See Also
    --------

    :func:`~radis.misc.arrays.is_sorted_backward`
    """
    for i in range(a.size - 1):
        if a[i + 1] < a[i]:
            return False
    return True


@numba.njit
def is_sorted_backward(a):
    """Returns whether ``a`` is sorted in descending order.

    See Also
    --------

    :func:`~radis.misc.arrays.is_sorted`
    """
    for i in range(a.size - 1):
        if a[i + 1] > a[i]:
            return False
    return True


def bining(I, ymin=None, ymax=None, axis=1):
    """Averages a I multi-dimensional array (typically an image) along the y
    axis bining(I) corresponds to I.mean(axis=1) Nan are not taken into
    account.

    Parameters
    ----------

    I: numpy array
        intensity
    ymin: int [0-I.shape[1]]
        If None, 0 is used. Default ``None``.
    ymax: int [0-I.shape[1]]
        If None, I.shape[1] is used. Default ``None``.
    axis: int
        Default 1
    """
    I = np.array(I)  # convert to array in case it's a Pandas dataframe for instance
    if ymin is None:
        ymin = 0
    if ymax is None:
        ymax = I.shape[axis]
    if ymin < 0:
        print("Warning in bining. ymin ({0}) < 0".format(ymin))
    if ymax > I.shape[axis]:
        print(
            "Warning in bining. ymax ({0}) > yaxis length ({1})".format(
                ymax, I.shape[axis]
            )
        )
    return np.nanmean(I[:, ymin:ymax], axis=axis)


def count_nans(a):
    """Nan are good but only in India."""

    return np.isnan(a).sum()


def logspace(xmin, xmax, npoints):
    """Returns points from xmin to xmax regularly distributed on a logarithm
    space.

    Numpy's :py:func:`numpy.logspace` does the same from 10**xmin to 10**xmax
    """

    return np.logspace(np.log10(xmin), np.log10(xmax), npoints)


def numpy_add_at(LDM, k, l, m, I):
    """Add the linestrengths on the LDM grid.

    Uses the numpy implementation of :py:func:`~numpy.add.at`, which
    add arguments element-wise.

    Parameters
    ----------
    LDM : ndarray
        LDM grid
    k, l, m : array
        index
    I : array
        intensity to add

    Returns
    -------
    add: ndarray
        linestrengths distributed over the LDM grid

    Notes
    -----
    Cython version implemented in https://github.com/radis/radis/pull/234

    See Also
    --------
    :py:func:`numpy.add.at`

    """
    # print('Numpy add.at()')
    return np.add.at(LDM, (k, l, m), I)


## Cython functions:
try:
    import radis_cython_extensions as rcx
except (ModuleNotFoundError):
    add_at = numpy_add_at
    #  or use radis.misc.utils.NotInstalled() ?
else:
    add_at = rcx.add_at


if __name__ == "__main__":
    import pytest

    pytest.main(["../test/misc/test_arrays.py", "-s"])  # -s for showing console output



# -*- coding: utf-8 -*-
"""
Created on Tue Feb  2 13:51:40 2021

@author: erwan
"""


import pytest

from radis.io.hitemp import HITEMP_MOLECULES, HITEMPDatabaseManager, fetch_hitemp
from radis.misc.config import getDatabankList


@pytest.mark.fast
def test_relevant_files_filter():
    from radis.io.hitemp import keep_only_relevant

    files = [
        "02_00000-00500_HITEMP2010.zip",
        "02_00500-00625_HITEMP2010.zip",
        "02_00625-00750_HITEMP2010.zip",
        "02_00750-01000_HITEMP2010.zip",
        "02_01000-01500_HITEMP2010.zip",
        "02_01500-02000_HITEMP2010.zip",
        "02_02000-02125_HITEMP2010.zip",
        "02_02125-02250_HITEMP2010.zip",
        "02_02250-02500_HITEMP2010.zip",
        "02_02500-03000_HITEMP2010.zip",
        "02_03000-03250_HITEMP2010.zip",
        "02_03250-03500_HITEMP2010.zip",
        "02_03500-03750_HITEMP2010.zip",
        "02_03750-04000_HITEMP2010.zip",
        "02_04000-04500_HITEMP2010.zip",
        "02_04500-05000_HITEMP2010.zip",
        "02_05000-05500_HITEMP2010.zip",
        "02_05500-06000_HITEMP2010.zip",
        "02_06000-06500_HITEMP2010.zip",
        "02_06500-12785_HITEMP2010.zip",
    ]

    assert keep_only_relevant(files)[0] == files
    assert keep_only_relevant(files, wavenum_max=300)[0] == [
        "02_00000-00500_HITEMP2010.zip"
    ]
    assert keep_only_relevant(files, wavenum_min=7000)[0] == [
        "02_06500-12785_HITEMP2010.zip"
    ]
    assert keep_only_relevant(files, wavenum_min=600, wavenum_max=800)[0] == [
        "02_00500-00625_HITEMP2010.zip",
        "02_00625-00750_HITEMP2010.zip",
        "02_00750-01000_HITEMP2010.zip",
    ]


@pytest.mark.needs_connection
def test_fetch_hitemp_OH(verbose=True, *args, **kwargs):
    """Test proper download of HITEMP OH database.

    Good to test fetch_hitemp.
    ``13_HITEMP2020.par.bz2`` is only 900 kb so it can be run on online
    tests without burning the planet 🌳

    ⚠️ if using the default `chunksize=100000`, it uncompresses in one pass
    (only 57k lines for OH) and we cannot test that appending to the same HDF5
    works. So here we use a smaller chunksize of 20,000.
    .

    """

    df = fetch_hitemp("OH", cache="regen", chunksize=20000, verbose=3 * verbose)

    assert "HITEMP-OH" in getDatabankList()

    assert len(df) == 57019

    # Load again and make sure it works (ex: metadata properly loaded etc.):
    fetch_hitemp("OH")


@pytest.mark.needs_connection
def test_fetch_hitemp_partial_download_CO2(verbose=True, *args, **kwargs):
    """Test partial download of HITEMP CO2 database.

    Good to test fetch_hitemp.
    ``CO2-02_02500-03000_HITEMP2010.h5`` is only 20 MB so it can be run on online
    tests without burning the planet too much 🌳

    """
    from os.path import basename

    from radis.io.hitemp import keep_only_relevant

    wmin = 2700
    wmax = 3000

    # Check that we won't download all databae for a reduced range:
    # ... This is done at the HITEMPDatabaseManager level
    # ... unitest for this part:
    ldb = HITEMPDatabaseManager(
        name="HITEMP-CO2", molecule="CO2", local_databases="~/.radisdb/hitemp/"
    )
    local_files, _ = ldb.get_filenames(engine="auto")
    relevant_file, file_wmin, file_wmax = keep_only_relevant(
        local_files, wavenum_min=wmin, wavenum_max=wmax
    )

    assert len(relevant_file) == 1
    assert basename(relevant_file[0]) == "CO2-02_02500-03000_HITEMP2010.h5"
    assert file_wmin == 2500
    assert file_wmax == 3000

    # Now run the full function :

    df, local_files = fetch_hitemp(
        "CO2",
        load_wavenum_min=2700,
        load_wavenum_max=3000,
        verbose=3,
        return_local_path=True,
    )

    assert "HITEMP-CO2" in getDatabankList()

    assert len(local_files) == 1
    assert basename(local_files[0]) == "CO2-02_02500-03000_HITEMP2010.h5"


@pytest.mark.needs_connection
@pytest.mark.download_large_databases
@pytest.mark.parametrize("molecule", [mol for mol in HITEMP_MOLECULES])
def test_fetch_hitemp_all_molecules(molecule, verbose=True, *args, **kwargs):
    """Test fetch HITEMP for all molecules whose download URL is available.

    ..warning::
        this downloads gigabytes of data. It is unselected by default by Pytest
        (see radis/setup.cfg)

    The bz2 compression factor gives about 17 MB / million lines :

    - OH (57 k lines) is only 900 kb
    - CO (0.7 M lines) is only ~14 Mb
    - CH4 (114 M lines) is 435 MB

    If it fails, check the databases downloaded in ~/.radisdb



    Notes
    -----

    Performance tests of chunksize tested on CO :
        - chunksize=1000000  > 22s  , 1 iteration ~ 22s
        - chunksize=100000 > 18s,  , 1 iteration ~ 4s
        - chunksize=50000 > 19s   ,1 iteration ~ 2s
        - chunksize=1000 --> 90s  ,  1 iteration << 1s
    """

    df, local_files = fetch_hitemp(
        molecule,
        columns=["int", "wav"],
        verbose=verbose,
        return_local_path=True,
        engine="pytables",
    )

    assert f"HITEMP-{molecule}" in getDatabankList()

    ldb = HITEMPDatabaseManager(
        name=f"HITEMP-{molecule}",
        molecule=molecule,
        verbose=verbose,
        local_databases="~/.radisdb/hitemp/",
    )
    url, Nlines, _, _ = ldb.fetch_url_Nlines_wmin_wmax()

    assert len(df) == Nlines


@pytest.mark.needs_connection
def test_partial_loading(*args, **kwargs):
    """Assert that using partial loading of the database works

    Also check 'vaex' engine and ``radis.config["AUTO_UPDATE_DATABASE"]``"""

    wmin, wmax = 2500, 4500

    # First ensures that the database is wider than this (else test is irrelevant) :
    df = fetch_hitemp("OH", engine="pytables")
    assert df.wav.min() < wmin
    assert df.wav.max() > wmax

    # Now fetch with partial loading
    df = fetch_hitemp(
        "OH", load_wavenum_min=wmin, load_wavenum_max=wmax, engine="pytables"
    )
    assert df.wav.min() >= wmin
    assert df.wav.max() <= wmax

    # Test with isotope:
    wmin2 = 0
    wmax2 = 300
    df = fetch_hitemp(
        "OH",
        load_wavenum_min=wmin2,
        load_wavenum_max=wmax2,
        isotope="2",
        engine="pytables",
    )
    assert df.iso.unique() == 2
    df = fetch_hitemp(
        "OH",
        load_wavenum_min=wmin2,
        load_wavenum_max=wmax2,
        isotope="1,2",
        engine="pytables",
    )
    assert set(df.iso.unique()) == {1, 2}

    # Check vaex engine :
    import radis

    old_config = radis.config["AUTO_UPDATE_DATABASE"]
    try:
        radis.config["AUTO_UPDATE_DATABASE"] = True
        df = fetch_hitemp("OH", engine="vaex")
    finally:
        radis.config["AUTO_UPDATE_DATABASE"] = old_config
    assert df.wav.min() < wmin
    assert df.wav.max() > wmax

    # multiple isotope selection not implemetned with vaex
    with pytest.raises(NotImplementedError):
        fetch_hitemp(
            "OH",
            load_wavenum_min=wmin,
            load_wavenum_max=wmax,
            isotope="1,2,3",
            engine="vaex",
        )


@pytest.mark.needs_connection
def test_calc_hitemp_spectrum(*args, **kwargs):
    """
    Test direct loading of HDF5 files
    """

    from astropy import units as u

    from radis import calc_spectrum

    calc_spectrum(
        wavenum_min=2500 / u.cm,
        wavenum_max=4500 / u.cm,
        molecule="OH",
        Tgas=600,
        databank="hitemp",  # test by fetching directly
        verbose=False,
    )

    calc_spectrum(
        wavenum_min=2500 / u.cm,
        wavenum_max=4500 / u.cm,
        molecule="OH",
        Tgas=600,
        databank="HITEMP-OH",  # test by loading the downloaded database
        verbose=False,
    )

    return


@pytest.mark.needs_connection
def test_calc_hitemp_CO_noneq(verbose=True, *args, **kwargs):
    """Test proper download of HITEMP CO database.

    Good to test noneq calculations with direct-download of HITEMP.

    ``05_HITEMP2020.par.bz2`` is about 14 Mb. We still download it
    (once per machine) because it allows to compute & test noneq spectra,
    which failed with direct HITEMP download in RADIS 0.9.28.

    Approximate cost for the planet 🌳 :  ~ 14 gr CO2
    (hard to evaluate !).


    """

    from astropy import units as u

    from radis import calc_spectrum

    calc_spectrum(
        wavenum_min=2000 / u.cm,
        wavenum_max=2300 / u.cm,
        molecule="CO",
        isotope="1,2",
        Tvib=1500,
        Trot=300,
        databank="hitemp",  # test by fetching directly
    )

    # Recompute with (now) locally downloaded database [note : this failed on 0.9.28]
    calc_spectrum(
        wavenum_min=2000 / u.cm,
        wavenum_max=2300 / u.cm,
        molecule="CO",
        isotope="1,2",
        Tvib=1500,
        Trot=300,
        databank="HITEMP-CO",  # registered in ~/radis.json
    )


@pytest.mark.needs_connection
@pytest.mark.download_large_databases
def test_parse_hitemp_missing_labels_issue280(*args, **kwargs):
    """Test dtype problems resulting from missing labels

    Issue 280 https://github.com/radis/radis/issues/280"""

    df = fetch_hitemp(
        "CO2", load_wavenum_min=800, load_wavenum_max=1300, verbose=3, cache="regen"
    )
    print(df.dtypes["v3u"])  # Worked


if __name__ == "__main__":
    test_relevant_files_filter()
    test_fetch_hitemp_OH()
    test_partial_loading()
    test_calc_hitemp_CO_noneq()
    test_fetch_hitemp_partial_download_CO2()
    test_calc_hitemp_spectrum()
    test_fetch_hitemp_all_molecules("OH")
    test_fetch_hitemp_all_molecules("CO")
    test_fetch_hitemp_all_molecules("CO2", verbose=3)
    test_fetch_hitemp_all_molecules("N2O", verbose=3)
    test_fetch_hitemp_all_molecules("NO", verbose=3)



# -*- coding: utf-8 -*-
"""Functions to update :class:`~radis.spectrum.spectrum.Spectrum` objects with
new spectral quantities that can be derived from existing ones, or rescale
path_length or mole_fraction, or add overpopulations.

Most of these are binded as methods to the Spectrum class, but stored here to
unload the spectrum.py file


-------------------------------------------------------------------------------
"""

from warnings import warn

import numpy as np
from numpy import exp
from numpy import log as ln

from radis.misc.basics import all_in, any_in, compare_lists
from radis.misc.debug import printdbg
from radis.spectrum.equations import calc_radiance
from radis.spectrum.utils import CONVOLUTED_QUANTITIES, NON_CONVOLUTED_QUANTITIES

# List of all spectral variables sorted by priority during recomputation
# (ex: first get abscoeff, then try to calculate emisscoeff, etc.)
ordered_keys = [
    "abscoeff",
    "emisscoeff",
    "absorbance",
    "radiance_noslit",
    "transmittance_noslit",
    "emissivity",
    "emissivity_noslit",
    "radiance",
    "transmittance",
]
"""list: List of all spectral variables sorted by priority during recomputation
(ex: first get abscoeff, then try to calculate emisscoeff, etc.)
"""

# ... variables that cannot be rescaled (or not implemented):
non_rescalable_keys = ["abscoeff_continuum", "emisscoeff_continuum"]
"""str:  variables that cannot be rescaled (or not implemented): """
# ... Check we have everyone (safety check!):
# ... if it fails here, then we may have added a new key without adding a scaling
# ... method. Explicitely add it in non_rescalableçkeys so an error is raised
# ... if trying to rescale a Spectrum that has such a quantity
assert (
    compare_lists(
        ordered_keys + non_rescalable_keys,
        CONVOLUTED_QUANTITIES + NON_CONVOLUTED_QUANTITIES,
        verbose=False,
    )
    == 1
)


def _build_update_graph(
    spec, optically_thin=None, equilibrium=None, path_length=None, no_change=False
):
    """Find inheritances properties (dependencies and equivalences) between all
    spectral variables based on the spectrum conditions (equilibrium, optically
    thin, known path length?)

    Parameters
    ----------

    spec: Spectrum
        a :class:`~radis.spectrum.spectrum.Spectrum` object

    Other Parameters
    ----------------

    optically_thin: boolean
        know whether the Spectrum should be considered optically thin to build
        the equivalence graph tree. If None, the value stored in the Spectrum is used.
        Default ``None``

    equilibrium: boolean
        know whether the Spectrum should be considered at equilibrium to build
        the equivalence graph tree. If None, the value stored in the Spectrum is used.
        Default ``None``

    path_length: boolean
        know whether the path length is given to build the equivalence graph tree.
        If None, ``path_length`` is looked up in the Spectrum condition. Default ``None``

    no_change: boolean
        if True, signals that we are somehow rescaling without changing path length
        nor mole fractions, i.e, all quantities can be recomputed from themselves...

    Returns
    -------

    derivation: dict
        {spectral_quantity: [list of combinations of spectral quantities needed to calculate it]}

    Examples
    --------

    to recompute a Spectrum under nonequilibrium, non optically thin case
    (note that all paths are not there yet)::

        {'abscoeff': [['absorbance']],
         'absorbance': [['transmittance_noslit'], ['abscoeff']],
         'emisscoeff': [['radiance_noslit', 'abscoeff']],
         'radiance_noslit': [['emisscoeff', 'abscoeff']],
         'transmittance_noslit': [['absorbance']]}

    a Spectrum under nonequilibrium, with optically thin case::

        {'abscoeff': [['absorbance']],
         'absorbance': [['transmittance_noslit'], ['abscoeff']],
         'emisscoeff': [['radiance_noslit']],
         'radiance_noslit': [['emisscoeff']],
         'transmittance_noslit': [['absorbance']]}

    a Spectrum under equilibrium (everything leads to everything)::

        {'abscoeff': [['absorbance'],
                      ['absorbance'],
                      ['emisscoeff'],
                      ['emissivity_noslit'],
                      ['transmittance'],
                      ['radiance'],
                      ['radiance_noslit'],
                      ['transmittance_noslit']],
         'absorbance': [['transmittance_noslit'],
                      ['abscoeff'],
                      ['abscoeff'],
                      ['emisscoeff'],
                      ['emissivity_noslit'],
                      ['transmittance'],
                      ['radiance'],
                      ['radiance_noslit'],
                      ['transmittance_noslit']],
         etc. }
    """

    # Get defaults
    if path_length is None:
        path_length = "path_length" in spec.conditions
    if optically_thin is None:
        optically_thin = spec.is_optically_thin()
    if equilibrium is None:
        # Read the Spectrum conditions. By default, use False.
        equilibrium = spec.conditions.get(
            "thermal_equilibrium", False
        )  # is_at_equilibrium()
    slit = (
        "slit_function" in spec.conditions
        and "slit_unit" in spec.conditions
        and "norm_by" in spec.conditions
    )

    all_keys = [
        "abscoeff",
        "absorbance",
        "emisscoeff",
        "emissivity_noslit",
        "transmittance",
        "radiance",
        "radiance_noslit",
        "transmittance_noslit",
    ]
    assert all_in(all_keys, CONVOLUTED_QUANTITIES + NON_CONVOLUTED_QUANTITIES)

    # Build edges of relationships
    derivation = {  # {keys, [list of keys]}
        "transmittance_noslit": [["absorbance"]],
        "absorbance": [["transmittance_noslit"]],
    }

    def derives_from(what, *from_keys):
        """Writes that quantity ``what`` can be infered by having all
        quantities ``from_keys``

        Examples
        --------

        Radiance can be infered from emisscoeff if optically thin::

            derives_from('radiance_noslit', 'emisscoeff')

        Else abscoeff would also be needed::

            derives_from('emisscoeff', ['radiance_noslit', 'abscoeff'])
        """
        for k in from_keys:
            if isinstance(k, str):
                k = [k]
            try:
                derivation[what].append(k)
            except KeyError:
                derivation[what] = [k]

    # Build more equivalence relationships if path_length is given
    # ------------------------------------------------------------
    # TODO: complete that, list is not exhaustive yet.
    # Duplicates are removed afterwards anyway
    #
    # Note for Developers: all derives_from relationship should correspond to a
    # rescale method that was implemented. Only the developer can know that!
    # If a rescaled relationship is implemetend but not added here it wont be
    # used by the code when trying to add all quantities. If a relationship is
    # added here but not implemented it will crash during rescale (and explain why)

    # no change: case where we are rescaling without changing path length nor
    # mole fractions, i.e, all initial quantities can be recomputed from themselves...
    if no_change:
        for k in all_keys:
            derives_from(k, [k])

    # Deal with case where we know path_length:
    if path_length:
        derives_from("abscoeff", ["absorbance"])
        derives_from("absorbance", ["abscoeff"])
        if optically_thin:
            derives_from("radiance_noslit", ["emisscoeff"])
            derives_from("emisscoeff", ["radiance_noslit"])
        else:
            derives_from("radiance_noslit", ["emisscoeff", "abscoeff"])
            derives_from("emisscoeff", ["radiance_noslit", "abscoeff"])

    if slit:
        if __debug__:
            printdbg("... build_graph: slit given > convoluted keys can be recomputed")
        derives_from("radiance", ["radiance_noslit"])
        derives_from("transmittance", ["transmittance_noslit"])
        derives_from("emissivity", ["emissivity_noslit"])

    if equilibrium:
        if __debug__:
            printdbg("... build_graph: equilibrium > all keys derive from one")
        # Anything can be recomputed from anything
        for key in all_keys:
            if key in NON_CONVOLUTED_QUANTITIES:
                all_but_k = [[k] for k in all_keys if k != key]
                derives_from(key, *all_but_k)

    # ------------------------------------------------------------

    if __debug__:
        printdbg("_build_update_graph: dependence/equivalence tree:")
        printdbg(derivation)

    return derivation


def get_redundant(spec):
    """Returns a dictionary of all spectral quantities in spectrum and whether
    they are redundant.

    Examples
    --------

    ::

        redundant = get_redundant(spec)

    >>> {'abscoeff': False, 'emisscoeff': True, 'absorbance': True, 'radiance_noslit': False,
         'transmittance_noslit': True}
    """

    # Get all spectral quantities that derive from existing ones
    derivation_graph = _build_update_graph(spec)

    activated = dict().fromkeys(ordered_keys, False)
    for k in spec.get_vars():
        activated[k] = True
    redundant = dict().fromkeys(ordered_keys, False)  # nothing redundant by default

    # Parse graph
    for key in ordered_keys[::-1]:  # roots
        if key in derivation_graph:
            for from_keys in derivation_graph[key]:
                #                if [key] == from_keys:
                #                    # that you can be recomputed from yourself doesnt make you redundant
                #                    continue
                if all([activated[k] and not redundant[k] for k in from_keys]):
                    redundant[key] = True  # declare redundant
                    continue
        else:
            if key not in spec.get_vars():
                del redundant[key]

    return redundant


def _path_is_complete(list_of_keys, computed_keys):
    return all([computed_keys[k] for k in list_of_keys])


def get_reachable(spec):  # , derivation_graph):
    """Get the list of all quantities that can be derived from current
    available quantities, based on given spec conditions.

    Parameters
    ----------

    spec: Spectrum
        a :class:`~radis.spectrum.spectrum.Spectrum` object

    Returns
    -------

    reachable: list
        list of quantities that can be calculated from available information

    Notes
    -----

    Algorithm::

        for all quantities, starting from the last:
            for all possible ways to compute them
                if valid, add quantity to reachable list, and restart
                else, continue
    """

    # Get inheritance links based on Spectrum conditions (equilibrium, optically thin, etc.)
    derivation_graph = _build_update_graph(spec)

    #    activated = dict().fromkeys(ordered_keys, False)
    reachable = dict().fromkeys(ordered_keys, False)
    for k in spec.get_vars():
        reachable[k] = True

    # Parse graph
    restart = True
    while restart:
        restart = False
        for key in ordered_keys[::-1]:  # roots
            if key in derivation_graph:
                # all different ways to compute this value
                for from_keys in derivation_graph[key]:
                    # if all are reachable then we can reach this new value
                    if _path_is_complete(from_keys, reachable):
                        if not reachable[key]:
                            reachable[key] = True
                            restart = True  # status changed -> restart?

    return reachable


# , derivation_graph):
def get_recompute(spec, wanted, no_change=False, true_path_length=None):
    """Get the list of all quantities that need to be recomputed to get the
    ``wanted`` quantities based on given spec conditions (does not recompute
    yet!)

    Parameters
    ----------

    spec: Spectrum
        a :class:`~radis.spectrum.spectrum.Spectrum` object

    wanted: list
        list of quantities to recompute


    Other Parameters
    ----------------

    no_change: boolean
        if True, signals that we are somehow rescaling without changing path length
        nor mole fractions, i.e, all quantities can be recomputed from themselves...

    true_path_length: boolean
        know whether the path length is given to build the equivalence graph tree.
        If None, ``path_length`` is looked up in the Spectrum condition. Default ``None``

    Returns
    -------

    recompute: list
        list of quantities needed

    Notes
    -----

    Algorithm::

        for all quantities:
            for all possible ways to compute them
                if valid, add quantity to reachable list, and restart
                else, continue
    """

    # Get inheritance links based on Spectrum conditions (equilibrium, optically thin, etc.)
    derivation_graph = _build_update_graph(
        spec, path_length=true_path_length, no_change=no_change
    )

    #    activated = dict().fromkeys(ordered_keys, False)
    # Store two dictionaries, that characterize, at a given instant, all quantities
    # that we had to recompute, and all quantities that can be recomputed from these
    recompute = dict().fromkeys(ordered_keys, False)
    for k in spec.get_vars():  # start from all quantities we have
        recompute[k] = True
    for k in wanted:  # add all quantities we want
        recompute[k] = True
    #    reachable = dict().fromkeys(ordered_keys, False)
    #    for k in spec.get_vars():   # start from all quantities we have
    #        reachable[k] = True

    def parse_tree(recompute):
        for key in ordered_keys:
            if key in wanted:  # find a way to recompute it:
                can_recompute_key = False
                # all different ways to compute this value
                for from_keys in derivation_graph[key]:
                    # if all are recomputed already they we can recompute this new value
                    if _path_is_complete(from_keys, recompute):
                        # we can reach this quantity
                        recompute[key] = True
                        can_recompute_key = True
                        break
                if not can_recompute_key:
                    if __debug__:
                        printdbg(
                            "... get_recompute: Can't recompute {0} with current keys: {1}. Finding something else".format(
                                key, [k for k in from_keys if recompute[k]]
                            )
                        )
                    #                    # cant recompute this quantity. Let's force recomputation
                    #                    def get_best_path():
                    #                        new_recompute_set = recompute.copy()
                    #                        for k in ordered_keys:
                    #                            # let's add a new quantity to recompute
                    #                            if new_recompute_set[k]:
                    #                                # we already have this quantity
                    #                                continue
                    #                            elif not any([k in from_keys for from_keys in derivation_graph[key]]):
                    #                                # this quantity (k) doesnt participate in calculating from_keys
                    #                                continue
                    #                            else:
                    #                                # let's calculate it
                    #
                    # cant recompute this quantity. Let's force recomputation
                    # of a given path. We'll arbitrary use the one will the fewer
                    # amount of not already recomputed quantities
                    score_non_recomputed_per_path = {}
                    for from_keys in derivation_graph[key]:
                        score = sum([not recompute[k] for k in from_keys])
                        # if several paths have same score the last will be chosen:
                        score_non_recomputed_per_path[score] = from_keys
                    # get the path with the minimum of values to recompute:
                    min_path = score_non_recomputed_per_path[
                        min(score_non_recomputed_per_path)
                    ]
                    # Add all these variables to the recompute list, and restart
                    for k in min_path:
                        recompute[k] = True
                        if __debug__:
                            printdbg(
                                "... get_recompute: Added new quantity to recompute list:",
                                k,
                            )
                    return recompute, True
        # reached the end with no change, no need to restart
        return recompute, False

    # Parse graph
    restart = True
    while restart:
        recompute, restart = parse_tree(recompute)

    recompute = [k for k in recompute if recompute[k]]

    if __debug__:
        printdbg("... get_recompute: List of quantities to recompute: ", recompute)

    return recompute


def update(
    spec,
    quantity="all",
    optically_thin="default",
    assume_equilibrium="default",
    verbose=True,
):
    """Calculate missing quantities that can be derived from the current
    quantities and conditions.

    e.g: if path_length and emisscoeff are given, radiance_noslit can be recalculated
    if in an optically thin configuration, else if abscoeff is also given


    Parameters
    ----------

    spec: Spectrum

    quantity: str, ``'same'``, ``'all'``, or list of str
        name of the spectral quantity to recompute. If ``'same'``, only the quantities
        in the Spectrum are recomputed. If ``'all'``, then all quantities that can
        be derived are recomputed. Default ``'all'``. See
        :py:data:`~radis.spectrum.utils.CONVOLUTED_QUANTITIES`
        and :py:data:`~radis.spectrum.utils.NON_CONVOLUTED_QUANTITIES`

    optically_thin: True, False, or ``'default'``
        determines whether to calculate radiance with or without self absorption.
        If 'default', the value is determined from the self_absorption key
        in Spectrum.conditions. If not given, False is taken. Default 'default'
        Also updates the self_absorption value in conditions (creates it if
        doesnt exist)

    Other Parameters
    ----------------

    assume_equilibrium: boolean
        if ``True``, only absorption coefficient ``abscoeff`` is recomputed
        and all values are derived from a calculation under equilibrium,
        using Kirchoff's Law. If ``default``, use the value stored in
        Spectrum conditions[``thermal_equilibrium``], else use ``False``.
    """

    # Check inputs
    # ------------

    # Get path length
    if "path_length" in list(spec.conditions.keys()):
        path_length = spec.conditions["path_length"]
        true_path_length = True
    else:
        path_length = 1  # some stuff can still be updated.
        true_path_length = False

    # Update thermal equilibrium
    if assume_equilibrium not in [True, False, "default"]:
        raise ValueError("assume_equilibrium must be one of True, False, 'default'")
    if assume_equilibrium == "default":
        assume_equilibrium = spec.conditions.get("thermal_equilibrium", False)

    # Update optically thin
    if optically_thin not in [True, False, "default"]:
        raise ValueError("optically_thin must be one of True, False, 'default'")
    if optically_thin == "default":
        if "self_absorption" in list(spec.conditions.keys()):
            optically_thin = not spec.conditions["self_absorption"]
        else:
            optically_thin = False
    old_self_absorption = spec.conditions.get("self_absorption")
    if old_self_absorption != (not optically_thin):
        spec.conditions["self_absorption"] = not optically_thin
        if verbose:
            print(("self absorption set to:", spec.conditions["self_absorption"]))

    initial = spec.get_vars()

    # This is where everything happens:
    # ---------------------------------
    _recalculate(
        spec,
        quantity,
        # same path length (absolute value can matter)
        path_length,
        path_length,
        # same mole fractions (only the ratio matters)
        1,
        1,
        true_path_length=true_path_length,
        assume_equilibrium=assume_equilibrium,
        verbose=verbose,
    )

    # Output
    # ------

    # Get list of new quantities
    new_q = [k for k in spec.get_vars() if k not in initial]
    if verbose:
        if len(new_q) > 0:
            print(("{0} new quantities added: {1}".format(spec.get_name(), new_q)))

    # Final checks
    for k in new_q:
        # make sure units exist
        if not k in spec.units:
            raise ValueError("{0} added but unit is unknown".format(k))


# Rescale functions

# ... absorption coefficient


def rescale_abscoeff(
    spec,
    rescaled,
    initial,
    old_mole_fraction,
    new_mole_fraction,
    old_path_length,
    wunit,
    units,
    extra,
    true_path_length,
):
    """

    Parameters
    ----------

    spec: Spectrum

    old_path_length: float
        path length in cm

    """

    unit = None

    # case where we recomputed it already (somehow... ex: no_change signaled)
    if "abscoeff" in rescaled:
        if __debug__:
            printdbg("... rescale: abscoeff was scaled already")
        assert "abscoeff" in units
        return rescaled, units

    # First get initial abscoeff
    # ---------------------
    if "abscoeff" in initial:
        _, abscoeff_init = spec.get("abscoeff", wunit=wunit)
    elif "absorbance" in initial and true_path_length:  # aka: true path_lengths given
        if __debug__:
            printdbg("... rescale: abscoeff k1 = A/L1")
        _, A = spec.get("absorbance", wunit=wunit)
        abscoeff_init = A / old_path_length  # recalculate initial
        unit = "cm-1"
    elif "transmittance_noslit" in initial and true_path_length:
        if __debug__:
            printdbg("... rescale: abscoeff k1 = -ln(T1)/L1")
        # Get abscoeff from transmittance
        _, T1 = spec.get("transmittance_noslit", wunit=wunit)

        # We'll have a problem if the spectrum is optically thick
        b = T1 == 0  # no transmittance: optically thick mask
        if b.sum() > 0:
            msg = "Transmittance is saturated. Can't infer abscoeff. Please give absorbance"
            if "abscoeff" in extra:  # cant calculate this one but let it go
                abscoeff_init = None
                if __debug__:
                    printdbg(msg)
            else:
                raise ValueError(msg)

        # Else, let's calculate it
        abscoeff_init = -ln(T1) / old_path_length  # recalculate initial
        unit = "cm-1"
    elif "abscoeff" in extra:  # cant calculate this one but let it go
        abscoeff_init = None
    else:
        raise ValueError(
            "Can't rescale abscoeff if transmittance_noslit ({0}) ".format(
                "transmittance_noslit" in initial
            )
            + "or absorbance ({0}), ".format("absorbance" in initial)
            + "and true_path_length ({0}) ".format(true_path_length)
            + "are not given. Use optically_thin?"
        )

    # Then export rescaled value
    # --------------------
    if abscoeff_init is not None:
        if __debug__:
            printdbg("... rescale: abscoeff k2 = k1 * N2/N1")
        abscoeff = abscoeff_init * new_mole_fraction / old_mole_fraction  # rescale
        rescaled["abscoeff"] = abscoeff
    if unit is not None:
        units["abscoeff"] = unit

    return rescaled, units


# ... all, if equilibrium and abscoeff was rescaled


def _recompute_all_at_equilibrium(
    spec, rescaled, wavenumber, Tgas, new_path_length, true_path_length, units
):
    """

    Parameters
    ----------

    rescaled: dict
        abscoeff must be rescaled already
    """

    def get_unit_radiance():
        return spec.units.get("radiance_noslit", "mW/cm2/sr/nm")

    def get_unit_emisscoeff(unit_radiance):
        if "/cm2" in unit_radiance:
            return unit_radiance.replace("/cm2", "/cm3")
        else:
            return unit_radiance + "/cm"  # will be simplified by Pint afterwards

    assert true_path_length

    abscoeff = rescaled["abscoeff"]
    path_length = new_path_length

    absorbance = abscoeff * path_length

    # Generate output quantities
    transmittance_noslit = exp(-absorbance)
    emissivity_noslit = 1 - transmittance_noslit
    radiance_noslit = calc_radiance(
        wavenumber, emissivity_noslit, Tgas, unit=get_unit_radiance()
    )
    b = transmittance_noslit == 1  # optically thin mask
    emisscoeff = np.empty_like(abscoeff)
    emisscoeff[b] = radiance_noslit[b] / path_length  # recalculate (opt thin)
    emisscoeff[~b] = (
        radiance_noslit[~b] / (1 - transmittance_noslit[~b]) * abscoeff[~b]
    )  # recalculate (non opt thin)

    # ----------------------------------------------------------------------

    rescaled["absorbance"] = absorbance
    rescaled["transmittance_noslit"] = transmittance_noslit
    rescaled["emissivity_noslit"] = emissivity_noslit
    rescaled["radiance_noslit"] = radiance_noslit
    rescaled["emisscoeff"] = emisscoeff

    units["abscoeff"] = "cm-1"
    units["absorbance"] = ""
    units["transmittance_noslit"] = ""
    units["emissivity_noslit"] = ""
    units["radiance_noslit"] = get_unit_radiance()
    units["emisscoeff"] = get_unit_emisscoeff(units["radiance_noslit"])

    return rescaled, units


# ... emission coefficient


def rescale_emisscoeff(
    spec,
    rescaled,
    initial,
    old_mole_fraction,
    new_mole_fraction,
    old_path_length,
    optically_thin,
    wunit,
    units,
    extra,
    true_path_length,
):
    """

    Parameters
    ----------

    spec: Spectrum
    """

    unit = None

    def get_unit(unit_radiance):
        if "/cm2" in unit_radiance:
            return unit_radiance.replace("/cm2", "/cm3")
        else:
            return unit_radiance + "/cm"  # will be simplified by Pint afterwards

    # case where we recomputed it already (somehow... ex: no_change signaled)
    if "emisscoeff" in rescaled:
        if __debug__:
            printdbg("... rescale: emisscoeff was scaled already")
        assert "emisscoeff" in units
        return rescaled, units

    # Firt get initial emisscoeff j1
    # -------------------

    if "emisscoeff" in initial:
        if __debug__:
            printdbg("... rescale: emisscoeff j1 = j1")
        _, emisscoeff_init = spec.get(
            "emisscoeff", wunit=wunit, Iunit=units["emisscoeff"]
        )

    elif "radiance_noslit" in initial and true_path_length and optically_thin:
        if __debug__:
            printdbg("... rescale: emisscoeff j1 = I1/L1")
        _, I = spec.get("radiance_noslit", wunit=wunit, Iunit=units["radiance_noslit"])
        emisscoeff_init = I / old_path_length  # recalculate initial
        unit = get_unit(units["radiance_noslit"])

    elif "radiance_noslit" in initial and true_path_length and "abscoeff" in initial:
        if __debug__:
            printdbg("... rescale: emisscoeff j1 = k1*I1/(1-exp(-k1*L1))")
        # get emisscoeff from (initial) abscoeff and (initial) radiance
        _, I = spec.get("radiance_noslit", wunit=wunit, Iunit=units["radiance_noslit"])
        _, k = spec.get("abscoeff", wunit=wunit, Iunit=units["abscoeff"])

        # Recalculate in the optically thin range (T=1) and elsewhere
        b = k == 0  # optically thin mask
        emisscoeff_init = np.empty_like(k)
        # ... optically thin case
        # recalculate (opt thin)
        emisscoeff_init[b] = I[b] / old_path_length

        # ... non optically thin case:
        # ... recalculate transmittance from abscoeff
        T_b = exp(-k[~b] * old_path_length)  # recalculate initial
        # ... and solve the RTE on an homogeneous slab
        # recalculate (non opt thin)
        emisscoeff_init[~b] = k[~b] * I[~b] / (1 - T_b)
        unit = get_unit(units["radiance_noslit"])

    elif (
        "radiance_noslit" in initial
        and true_path_length
        and "transmittance_noslit" in initial
    ):
        if __debug__:
            printdbg("... rescale: emisscoeff j1 = k1*I1/(1-T1)")
        # get emisscoeff from (initial) transmittance and (initial) radiance
        _, I = spec.get("radiance_noslit", wunit=wunit, Iunit=units["radiance_noslit"])
        _, T = spec.get(
            "transmittance_noslit", wunit=wunit, Iunit=units["transmittance_noslit"]
        )

        # Recalculate in the optically thin range (T=1) and elsewhere
        b = T == 1  # optically thin mask
        emisscoeff_init = np.empty_like(T)
        # ... optically thin case
        # recalculate (opt thin)
        emisscoeff_init[b] = I[b] / old_path_length

        # ... non optically thin case:
        # ... recalculate abscoeff from transmittance
        T_b = T[~b]
        k_b = -ln(T_b) / old_path_length
        # ... and solve the RTE on an homogeneous slab
        # recalculate (non opt thin)
        emisscoeff_init[~b] = k_b * I[~b] / (1 - T_b)
        unit = get_unit(units["radiance_noslit"])

    else:
        if optically_thin:
            msg = "Can't calculate emisscoeff if true path_length ({0})".format(
                true_path_length
            ) + "and initial radiance_noslit ({0}) are not all given".format(
                "radiance_noslit" in initial
            )
            if "emisscoeff" in extra:  # cant calculate this one but let it go
                emisscoeff_init = None
                if __debug__:
                    printdbg(msg)
            else:
                raise ValueError(msg)
        else:
            msg = (
                "Trying to get the emission coefficient (emisscoeff) in non optically "
                + "thin case. True path_length ({0}), radiance_noslit ({1}) ".format(
                    true_path_length, "radiance_noslit" in initial
                )
                + "and abscoeff ({0}) are needed but not all given. ".format(
                    "abscoeff" in initial
                )
                + "Try optically_thin? See known Spectrum conditions with "
                + "print(Spectrum)"
            )
            if "emisscoeff" in extra:  # cant calculate this one but let it go
                emisscoeff_init = None
                if __debug__:
                    printdbg(msg)
            else:
                raise ValueError(msg)

    # Then rescale and export
    # -----------------------
    if emisscoeff_init is not None:
        if __debug__:
            printdbg("... rescale: emisscoeff j2 = j1 * N2/N1")
        # Now rescale for mole fractions
        emisscoeff = (
            emisscoeff_init * new_mole_fraction / old_mole_fraction
        )  # rescale x
        rescaled["emisscoeff"] = emisscoeff
    if unit is not None:
        units["emisscoeff"] = unit

    return rescaled, units


# ... absorbance


def rescale_absorbance(
    spec,
    rescaled,
    initial,
    old_mole_fraction,
    new_mole_fraction,
    old_path_length,
    new_path_length,
    waveunit,
    units,
    extra,
    true_path_length,
):
    """

    Parameters
    ----------

    spec: Spectrum
    """

    unit = None

    # case where we recomputed it already (somehow... ex: no_change signaled)
    if "absorbance" in rescaled:
        if __debug__:
            printdbg("... rescale: absorbance was scaled already")
        assert "absorbance" in units
        return rescaled, units

    # Get scaled absorbance directly A1
    # ---------------------------------

    if "absorbance" in initial:
        if __debug__:
            printdbg("... rescale: absorbance A2 = A1*N2/N1*L2/L1")
        _, absorbance = spec.get(
            "absorbance", wunit=waveunit, Iunit=units["absorbance"]
        )
        absorbance *= new_mole_fraction / old_mole_fraction  # rescale x
        absorbance *= new_path_length / old_path_length  # rescale L
        unit = ""
    elif "abscoeff" in rescaled and true_path_length:
        if __debug__:
            printdbg("... rescale: absorbance A2 = j2*L2")
        abscoeff = rescaled["abscoeff"]  # x already scaled
        absorbance = abscoeff * new_path_length  # calculate L
        unit = ""
    elif "transmittance_noslit" in initial and true_path_length:
        if __debug__:
            printdbg("... rescale: absorbance A2 = -ln(T1)*N2/N1*L2/L1")
        # Get absorbance from transmittance
        _, T1 = spec.get("transmittance_noslit", wunit=waveunit)

        # We'll have a problem if the spectrum is optically thick
        b = T1 == 0  # no transmittance: optically thick mask
        if b.sum() > 0:
            msg = "Transmittance is saturated. Can't calculate absorbance"
            if "absorbance" in extra:  # cant calculate this one but let it go
                absorbance = None
                if __debug__:
                    printdbg(msg)
            else:
                raise ValueError(msg)

        # Else, let's calculate it
        absorbance = -ln(T1)
        absorbance *= new_mole_fraction / old_mole_fraction  # rescale x
        absorbance *= new_path_length / old_path_length  # rescale L
        unit = ""
    else:
        msg = (
            "Cant recalculate absorbance if scaled absoeff "
            + "({0}) and true path_length ({1}) are not given".format(
                "abscoeff" in rescaled, true_path_length
            )
        )
        if "absorbance" in extra:  # cant calculate this one but let it go
            absorbance = None
            if __debug__:
                printdbg(msg)
        else:
            raise ValueError(msg)

    # Then rescale and export
    # -----------------------

    # Export rescaled value
    if absorbance is not None:
        rescaled["absorbance"] = absorbance
    if unit is not None:
        units["absorbance"] = unit

    return rescaled, units


# ... transmittance
def rescale_transmittance_noslit(
    spec,
    rescaled,
    initial,
    old_mole_fraction,
    new_mole_fraction,
    old_path_length,
    new_path_length,
    waveunit,
    units,
    extra,
    true_path_length,
):
    """

    Parameters
    ----------

    spec: Spectrum
    """

    unit = None

    def get_unit():
        return ""

    # case where we recomputed it already (somehow... ex: no_change signaled)
    if "transmittance_noslit" in rescaled:
        if __debug__:
            printdbg("... rescale: transmittance_noslit was scaled already")
        assert "transmittance_noslit" in units
        return rescaled, units

    # Calculate rescaled value directly
    # ---------------------------------

    # Rescale
    if "absorbance" in rescaled:
        if __debug__:
            printdbg("... rescale: transmittance_noslit T2 = exp(-A2)")
        absorbance = rescaled["absorbance"]  # x and L already scaled
        transmittance_noslit = exp(-absorbance)  # recalculate
        unit = get_unit()
    elif "abscoeff" in rescaled and true_path_length:
        if __debug__:
            printdbg("... rescale: transmittance_noslit T2 = exp(-j2*L2)")
        abscoeff = rescaled["abscoeff"]  # x already scaled
        absorbance = abscoeff * new_path_length  # calculate
        transmittance_noslit = exp(-absorbance)  # recalculate
        unit = get_unit()
    elif "transmittance_noslit" in initial:
        if __debug__:
            printdbg(
                "... rescale: transmittance_noslit T2 = "
                + "exp( ln(T1) * N2/N1 * L2/L1)"
            )
        # get transmittance from initial transmittance
        _, T1 = spec.get(
            "transmittance_noslit", wunit=waveunit, Iunit=units["transmittance_noslit"]
        )

        # We'll have a problem if the spectrum is optically thick
        b = T1 == 0  # optically thick mask
        if b.sum() > 0 and (
            new_mole_fraction < old_mole_fraction or new_path_length < old_path_length
        ):
            # decreasing mole fractions/ path length could increase the transmittance
            # but this information was lost in the saturation
            msg = "Transmittance is saturated. Cant rescale. Please give absorbance"
            if "transmittance_noslit" in extra:  # cant calculate this one but let it go
                transmittance_noslit = None
                if __debug__:
                    printdbg(msg)
            else:
                raise ValueError(msg)
        # Else, just get absorbance
        absorbance = -ln(T1)
        absorbance *= new_mole_fraction / old_mole_fraction  # rescale x
        absorbance *= new_path_length / old_path_length  # rescale L
        transmittance_noslit = exp(-absorbance)
    else:
        msg = "Missing data to rescale transmittance. Expected scaled absorbance ({0})".format(
            "absorbance" in rescaled
        )
        #        +' or scaled abscoeff ({0}) and true_path_length ({1})'.format(
        #                'abscoeff' in rescaled, true_path_length)
        if "transmittance_noslit" in extra:  # cant calculate this one but let it go
            transmittance_noslit = None
            if __debug__:
                printdbg(msg)
        else:
            raise ValueError(msg)

    # Export rescaled value
    if transmittance_noslit is not None:
        rescaled["transmittance_noslit"] = transmittance_noslit
    if unit is not None:
        units["transmittance_noslit"] = unit

    return rescaled, units


def rescale_transmittance(
    spec,
    rescaled,
    initial,
    old_mole_fraction,
    new_mole_fraction,
    old_path_length,
    new_path_length,
    waveunit,
    units,
    extra,
):
    """

    Parameters
    ----------

    spec: Spectrum
    """

    #    unit = None
    apply_slit = False
    #    def get_unit():
    #        return '1'

    # case where we recomputed it already (somehow... ex: no_change signaled)
    if "transmittance" in rescaled:
        if __debug__:
            printdbg("... rescale: transmittance was scaled already")
        assert "transmittance" in units
        return rescaled, units, apply_slit

    if "transmittance_noslit" in rescaled:
        apply_slit = True
    else:
        raise NotImplementedError("rescale transmittance not implemented")

    return rescaled, units, apply_slit


# ... radiance_noslit


def rescale_radiance_noslit(
    spec,
    rescaled,
    initial,
    old_mole_fraction,
    new_mole_fraction,
    old_path_length,
    new_path_length,
    optically_thin,
    waveunit,
    units,
    extra,
    true_path_length,
):
    """

    Parameters
    ----------

    spec: Spectrum
    """

    unit = None

    def get_radiance_unit(unit_emisscoeff):
        """get radiance_noslit unit from emisscoeff unit."""
        if "/cm3" in unit_emisscoeff:
            return unit_emisscoeff.replace("/cm3", "/cm2")
        else:
            return unit_emisscoeff + "*cm"

    # case where we recomputed it already (somehow... ex: no_change signaled)
    if "radiance_noslit" in rescaled:
        if __debug__:
            printdbg("... rescale: radiance_noslit was scaled already")
        assert "radiance_noslit" in units
        return rescaled, units

    # Rescale!
    if "emisscoeff" in rescaled and true_path_length and optically_thin:
        if __debug__:
            printdbg("... rescale: radiance_noslit I2 = j2 * L2 " + "(optically thin)")
        emisscoeff = rescaled["emisscoeff"]  # x already scaled
        radiance_noslit = emisscoeff * new_path_length  # recalculate L
        unit = get_radiance_unit(units["emisscoeff"])

    elif (
        "emisscoeff" in rescaled
        and "transmittance_noslit" in rescaled
        and "abscoeff" in rescaled
        and true_path_length
        and not optically_thin
    ):  # not optically thin
        if __debug__:
            printdbg("... rescale: radiance_noslit I2 = j2*(1-T2)/k2")
        emisscoeff = rescaled["emisscoeff"]  # x already scaled
        abscoeff = rescaled["abscoeff"]  # x already scaled
        # mole_fraction, path_length already scaled
        transmittance_noslit = rescaled["transmittance_noslit"]
        b = transmittance_noslit == 1  # optically thin mask
        radiance_noslit = np.empty_like(emisscoeff)  # calculate L
        radiance_noslit[~b] = (
            emisscoeff[~b] / abscoeff[~b] * (1 - transmittance_noslit[~b])
        )
        radiance_noslit[b] = emisscoeff[b] * new_path_length  # optically thin limit
        unit = get_radiance_unit(units["emisscoeff"])

    elif (
        "emisscoeff" in rescaled
        and "abscoeff" in rescaled
        and true_path_length
        and not optically_thin
    ):  # not optically thin
        if __debug__:
            printdbg("... rescale: radiance_noslit I2 = j2*(1-exp(-k2*L2))/k2")
        emisscoeff = rescaled["emisscoeff"]  # x already scaled
        abscoeff = rescaled["abscoeff"]  # x already scaled
        b = abscoeff == 0  # optically thin mask
        radiance_noslit = np.empty_like(emisscoeff)  # calculate
        radiance_noslit[~b] = (
            emisscoeff[~b] / abscoeff[~b] * (1 - exp(-abscoeff[~b] * new_path_length))
        )
        radiance_noslit[b] = emisscoeff[b] * new_path_length  # optically thin limit
        unit = get_radiance_unit(units["emisscoeff"])

    elif "radiance_noslit" in initial and optically_thin:
        if __debug__:
            printdbg(
                "... rescale: radiance_noslit I2 = I1*N2/N1*L2/L1 " + "(optically thin)"
            )
        _, radiance_noslit = spec.get(
            "radiance_noslit", wunit=waveunit, Iunit=units["radiance_noslit"]
        )
        radiance_noslit *= new_mole_fraction / old_mole_fraction  # rescale
        radiance_noslit *= new_path_length / old_path_length  # rescale

    else:
        if optically_thin:
            msg = (
                "Missing data to rescale radiance_noslit in "
                + "optically thin mode. You need at least initial "
                + "radiance_noslit ({0}), or scaled emission coefficient ({1}) ".format(
                    "radiance_noslit" in initial, "emisscoeff" in rescaled
                )
                + "and true path length ({0}).".format(true_path_length)
            )
            if "radiance_noslit" in extra:  # cant calculate this one but let it go
                radiance_noslit = None
                if __debug__:
                    printdbg(msg)
            else:
                raise ValueError(msg)
        else:
            msg = (
                "Missing data to recalculate radiance_noslit. You need at least "
                + "scaled emisscoeff ({0}), scaled transmittance_noslit ({1}) ".format(
                    "emisscoeff" in rescaled, "transmittance_noslit" in rescaled
                )
                + "scaled abscoeff ({0}) and true_path_length ({1}). ".format(
                    "abscoeff" in rescaled, true_path_length
                )
                + "Try in optically thin mode"
            )
            if "radiance_noslit" in extra:
                radiance_noslit = None
                if __debug__:
                    printdbg(msg)
            else:
                raise ValueError(msg)

    # Export rescaled value
    if radiance_noslit is not None:
        rescaled["radiance_noslit"] = radiance_noslit
    if unit is not None:
        units["radiance_noslit"] = unit

    return rescaled, units


# ... radiance


def rescale_radiance(
    spec,
    rescaled,
    initial,
    old_mole_fraction,
    new_mole_fraction,
    old_path_length,
    new_path_length,
    optically_thin,
    waveunit,
    units,
    extra,
    true_path_length,
):
    """

    Parameters
    ----------

    spec: Spectrum
    """

    apply_slit = False
    #    unit = None
    #    def get_unit(unit_emisscoeff):
    #        if '/cm3' in unit_emisscoeff:
    #            return unit_emisscoeff.replace('/cm3', '/cm2')
    #        else:
    #            return unit_emisscoeff + '*cm'

    # case where we recomputed it already (somehow... ex: no_change signaled)
    if "radiance" in rescaled:
        if __debug__:
            printdbg("... rescale: radiance was scaled already")
        assert "radiance" in units
        return rescaled, units, apply_slit

    # Rescale!
    if "radiance_noslit" in rescaled:
        apply_slit = True
    else:
        raise NotImplementedError("rescale radiance not implemented yet")

    return rescaled, units, apply_slit


# ... emissivity_noslit


def rescale_emissivity_noslit(spec, rescaled, units, extra, true_path_length):
    """

    Parameters
    ----------

    spec: Spectrum
    """

    # case where we recomputed it already (somehow... ex: no_change signaled)
    if "emissivity_noslit" in rescaled:
        if __debug__:
            printdbg("... rescale: emissivity_noslit was scaled already")
        assert "emissivity_noslit" in units
        return rescaled, units

    # Rescale!
    if "transmittance_noslit" in rescaled:
        if __debug__:
            printdbg("... rescale: emissivity_noslit e2 = 1 - T2")
        # transmittivity already scaled
        T2 = rescaled["transmittance_noslit"]
        emissivity_noslit = 1 - T2  # recalculate
    else:
        msg = "transmittance_noslit needed to recompute emissivity_noslit"
        if "emissivity_noslit" in extra:  # cant calculate this one but let it go
            emissivity_noslit = None
            if __debug__:
                printdbg(msg)
        else:
            raise ValueError(msg)

    # Export rescaled value
    if emissivity_noslit is not None:
        rescaled["emissivity_noslit"] = emissivity_noslit
        units["emissivity_noslit"] = ""

    return rescaled, units


def _recalculate(
    spec,
    quantity,
    new_path_length,
    old_path_length,
    new_mole_fraction,
    old_mole_fraction,
    true_path_length=True,
    verbose=True,
    assume_equilibrium=False,
):
    """General function to recalculate missing quantities. Used in
    :func:`~radis.spectrum.rescale.rescale_path_length`, :func:`~radis.spectrum.rescale.rescale_mole_fraction`
    and :func:`~radis.spectrum.rescale.update`.

    Determines with spectral quantities should be recomputed, then scales
    them solving the Radiative Transfer Equation in the process.


    Parameters
    ----------

    spec: Spectrum
        the Spectrum object to recompute

    quantity: str, ``'same'``, ``'all'``, or list of str
        name of the spectral quantity to recompute. If ``'same'``, only the quantities
        in the Spectrum are recomputed. If ``'all'``, then all quantities that can
        be derived are recomputed. See :py:data:`~radis.spectrum.utils.CONVOLUTED_QUANTITIES`
        and :py:data:`~radis.spectrum.utils.NON_CONVOLUTED_QUANTITIES`

    true_path_length: boolean
        if ``False``, only relative rescaling (new/old) is allowed. For instance,
        when you dont know the true path_lenth, rescaling absorbance
        with *= new_length/old_length is fine, but abscoeff*new_length is not
        Default ``True``

    Other Parameters
    ----------------

    assume_equilibrium: boolean
        if ``True``, only absorption coefficient ``abscoeff`` is recomputed
        and all values are derived from a calculation under equilibrium,
        using Kirchoff's Law. Default ``False``

    """

    optically_thin = spec.is_optically_thin()
    initial = spec.get_vars()  # quantities initially in spectrum
    if __debug__:
        printdbg("... rescale: optically_thin: {0}".format(optically_thin))
        printdbg("... rescale: initial quantities: {0}".format(initial))

    # Check that inputs are valid names
    _check_quantity = quantity
    if isinstance(quantity, list):
        _check_quantity = quantity
    else:
        _check_quantity = [quantity]
    for k in _check_quantity:
        assert k in CONVOLUTED_QUANTITIES + NON_CONVOLUTED_QUANTITIES + ["all", "same"]
    # ... make sure we're not trying to rescale a Spectrum that has non scalable
    # ... quantities
    if any_in(initial, non_rescalable_keys):
        raise NotImplementedError(
            "Trying to rescale a Spectrum that has non scalable "
            + "quantities: {0}".format([k for k in initial if k in non_rescalable_keys])
            + "Remove them manually, or implement the scaling method."
        )

    # Choose which values to recompute (and store them in the list `wanted`)
    # ----------
    if quantity == "all":  # quantities to recompute
        wanted = list(initial)  # start from the one we have (that also makes)
        # sure we dont delete anyone with the final validity
        # check "everyone is here"
        greedy = True
    elif quantity == "same":
        wanted = list(initial)
        greedy = False
    elif isinstance(quantity, str):
        wanted = [quantity]
        greedy = False
    elif isinstance(quantity, list):
        wanted = quantity
        greedy = False
    else:
        raise ValueError(
            "unexpected type for quantity: expected str, got "
            + "{0} ({1})".format(quantity, type(quantity))
        )
    rescaled = {}  # quantities rescaled

    # in greedy mode ('all'), choose to recompute all parameters that we can
    extra = []
    if greedy:
        # ... let's be greedy: recompute all possible quantities. The list of
        # all spectral quantities is calculated by parsing a tree in get_reachable
        reachable = get_reachable(spec)
        extra = [k for k, v in reachable.items() if v]
    wanted = set(wanted + extra)

    # There are two cases: either we are actually rescaling to another length /
    # mole fraction, or we are just updating() without changing length / mole fraction
    no_change = (
        new_mole_fraction == old_mole_fraction and new_path_length == old_path_length
    )

    # Quickly stop if no change
    if no_change and all_in(wanted, initial):
        if __debug__:
            printdbg("... rescale: no change")
        # Stop here
        return

    # list of quantities that are needed to recompute what we want
    # ... (we're just analysing how to compute them here, the actual calculation
    # ... will be done laters)
    try:
        recompute = get_recompute(
            spec, wanted, no_change, true_path_length=true_path_length
        )
    except KeyError as err:
        import sys

        print(sys.exc_info())
        raise KeyError(
            "Error in get_recompute (see above). Quantity `{0}` cannot be recomputed ".format(
                err.args[0]
            )
            + "from available quantities in Spectrum ({0}) with ".format(
                spec.get_vars()
            )
            + " conditions: optically thin ({0}), true_path_length ({1}), thermal equilibrium ({2})".format(
                optically_thin, true_path_length, assume_equilibrium
            )
            + ". Check how your equivalence tree is built: see rescale._build_update_graph()"
        )
    if assume_equilibrium:
        recompute.append("abscoeff")
    recompute = set(recompute)  # remove duplicates

    # Get units
    units = spec.units.copy()

    # Recompute!
    # ----------
    waveunit = spec.get_waveunit()  # keep all quantities in same waveunit
    apply_slit = False  # if True at the end re-apply slit

    # If no_change, just set everyone as rescaled already
    if no_change:
        for k in initial:
            rescaled[k] = spec.get(k)[1]  # note: creates a copy

    # Start with abscoeff

    if "abscoeff" in recompute:
        rescaled, units = rescale_abscoeff(
            spec,
            rescaled,
            initial,  # Todo: remove rescaled = ... Dict is mutable no?
            old_mole_fraction,
            new_mole_fraction,
            old_path_length,
            waveunit,
            units,
            extra,
            true_path_length,
        )

    if (
        assume_equilibrium
        and "Tgas" in spec.conditions
        and spec.conditions["Tgas"] != "N/A"
    ):
        # ... (dont forget Python will stop at evaluating the 2nd expression if False)
        assert "abscoeff" in rescaled
        if not spec.is_at_equilibrium():
            warn(
                "Rescaling with equilibrium assumption but Spectrum {0} does ".format(
                    spec.get_name()
                    + "not look at equilibrium. Check spectrum conditions"
                )
            )
        wavenumber = spec.get_wavenumber()
        Tgas = spec.conditions["Tgas"]
        rescaled, units = _recompute_all_at_equilibrium(
            spec, rescaled, wavenumber, Tgas, new_path_length, true_path_length, units
        )
        apply_slit = "radiance" in recompute or "transmittance" in recompute

    else:

        if "emisscoeff" in recompute:
            rescaled, units = rescale_emisscoeff(
                spec,
                rescaled,
                initial,
                old_mole_fraction,
                new_mole_fraction,
                old_path_length,
                optically_thin,
                waveunit,
                units,
                extra,
                true_path_length,
            )

        if "absorbance" in recompute:
            rescaled, units = rescale_absorbance(
                spec,
                rescaled,
                initial,
                old_mole_fraction,
                new_mole_fraction,
                old_path_length,
                new_path_length,
                waveunit,
                units,
                extra,
                true_path_length,
            )

        if "transmittance_noslit" in recompute:
            rescaled, units = rescale_transmittance_noslit(
                spec,
                rescaled,
                initial,
                old_mole_fraction,
                new_mole_fraction,
                old_path_length,
                new_path_length,
                waveunit,
                units,
                extra,
                true_path_length,
            )

        if "radiance_noslit" in recompute:
            rescaled, units = rescale_radiance_noslit(
                spec,
                rescaled,
                initial,
                old_mole_fraction,
                new_mole_fraction,
                old_path_length,
                new_path_length,
                optically_thin,
                waveunit,
                units,
                extra,
                true_path_length,
            )

        if "emissivity_noslit" in recompute:
            rescaled, units = rescale_emissivity_noslit(
                spec, rescaled, units, extra, true_path_length
            )

        if "radiance" in recompute:
            rescaled, units, slit_needed = rescale_radiance(
                spec,
                rescaled,
                initial,
                old_mole_fraction,
                new_mole_fraction,
                old_path_length,
                new_path_length,
                optically_thin,
                waveunit,
                units,
                extra,
                true_path_length,
            )
            apply_slit = apply_slit or slit_needed

        if "transmittance" in recompute:
            rescaled, units, slit_needed = rescale_transmittance(
                spec,
                rescaled,
                initial,
                old_mole_fraction,
                new_mole_fraction,
                old_path_length,
                new_path_length,
                waveunit,
                units,
                extra,
            )
            apply_slit = apply_slit or slit_needed

    # delete former convoluted value if apply_slit will be used (just to be sure
    # we arent keeping a non rescaled value if something goes wrong)
    if apply_slit:
        for k in list(spec._q.keys()):
            if k != "wavespace" and k in CONVOLUTED_QUANTITIES:
                del spec._q[k]
                del spec.units[k]

    # Save (only) the ones that we want, unless we want everything ('greedy')
    for q in rescaled:
        if q in wanted:  # or greedy:
            spec._q[q] = rescaled[q]

    # Update units
    for k, u in units.items():
        spec.units[k] = u

    # Reapply slit if needed
    # TODO: replace with directly convolving with slit stored in conditions
    # TODO: first, add an option to give arrays to apply_slit (see Issue #30)
    if apply_slit:
        if not (
            "slit_function" in spec.conditions
            and "slit_unit" in spec.conditions
            and "norm_by" in spec.conditions
        ):
            #            if 'transmittance' in extra and 'radiance' in extra:
            #                pass
            #            else:
            raise KeyError(
                "Slit is needed to recompute some quantities ({0}) ".format(wanted)
                + "but not all conditions are given among slit_function "
                + "({0}), slit_unit ({1}) and norm_by ({2})".format(
                    "slit_function" in spec.conditions,
                    "slit_unit" in spec.conditions,
                    "norm_by" in spec.conditions,
                )
            )
        else:
            slit_function = spec.conditions["slit_function"]
            slit_unit = spec.conditions["slit_unit"]
            norm_by = spec.conditions["norm_by"]
            try:
                shape = spec.conditions["slit_shape"]
            except KeyError:
                warn(
                    "Recomputing with slit_shape not given. Assuming a triangular slit.",
                    UserWarning,
                )
                shape = "triangular"
            spec.apply_slit(
                slit_function=slit_function,
                unit=slit_unit,
                shape=shape,
                norm_by=norm_by,
                verbose=verbose,
            )

    # Final checks

    # ... "everyone is here": check we didnt miss anyone
    rescaled_list = list(rescaled)
    # add the new quantities added by apply_slit
    rescaled_list = rescaled_list + [
        k for k in spec.get_vars() if k in CONVOLUTED_QUANTITIES
    ]
    for q in wanted:
        if not q in rescaled_list:
            raise AssertionError(
                "{0} could not be rescaled as wanted. ".format(q)
                + "The following properties were rescaled: {0}".format(rescaled_list)
            )
    # ... everyone was added in the Spectrum properly
    final_list = spec.get_vars()
    for q in wanted:
        if not q in final_list:
            raise AssertionError(
                "{0} is not in the final Spectrum. ".format(q)
                + "Rescaled spectrum contains: {0}".format(final_list)
            )
    # ... "everyone was rescaled": check we didnt scale only part of the spectrum
    for q in initial:
        if not q in rescaled_list:
            raise AssertionError(
                "{0} was initially in the Spectrum but was not ".format(q)
                + "rescaled. This can lead to error. Rescaled spectrum "
                + "contains: {0}".format(rescaled_list)
            )


def rescale_path_length(
    spec, new_path_length, old_path_length=None, inplace=False, force=False
):
    """Rescale spectrum to new path length. Starts from absorption coefficient
    and emission coefficient, and solves the RTE again for the new path length
    Convoluted values (with slit) are dropped in the process.

    Parameters
    ----------

    spec: Spectrum

    new_path_length: float
        new path length

    old_path_length: float, or None
        if None, current path length (conditions['path_length']) is used


    Other Parameters
    ----------------

    inplace: boolean
        if ``True``, modifies the Spectrum object directly. Else, returns
        a copy. Default ``False``.

    force: boolean
        if False, won't allow rescaling to 0 (not to loose information).
        Default ``False``

    Returns
    -------

    s_rescaled: Spectrum
        a rescaled Spectrum.
        if ``inplace=True``, then ``s`` has been rescaled already and
        ``s_rescaled`` is ``s``

    Notes
    -----

    Implementation:

        To deal with all the input cases, we first make a list of what has to
        be recomputed, and what has to be recalculated
    """

    if not inplace:
        spec = spec.copy()

    # Check inputs
    # ----------
    if old_path_length is not None:
        try:
            if spec.conditions["path_length"] != old_path_length:
                warn(
                    "path_length ({0}) doesnt match value given in conditions ({1})".format(
                        old_path_length, spec.conditions["path_length"]
                    )
                )
        except KeyError:  # path_length not defined
            pass
    else:
        try:
            old_path_length = spec.conditions["path_length"]
        except KeyError:
            raise KeyError(
                "path_length has to be defined in conditions (or use"
                + " `from_path_length`)"
            )

    if new_path_length < 0 and not force:
        raise ValueError("path_length cannot be negative")
    if new_path_length == 0 and not force:
        raise ValueError(
            "Rescaling to 0 will loose information. Choose force " "= True"
        )
    for q in ["transmittance", "radiance"]:
        qns = q + "_noslit"
        qties = spec.get_vars()
        if q in qties and qns not in qties and not force:
            raise KeyError(
                "Cant rescale {0} if {1} not stored. ".format(q, qns)
                + " Use force=True to rescale anyway. {0}".format(q)
                + " will be deleted"
            )

    # Rescale
    assume_equilibrium = spec.conditions.get("thermal_equilibrium", False)
    _recalculate(
        spec,
        "same",
        new_path_length,
        old_path_length,
        1,
        1,
        assume_equilibrium=assume_equilibrium,
    )

    # Update conditions
    spec.conditions["path_length"] = new_path_length

    return spec


def rescale_mole_fraction(
    spec,
    new_mole_fraction,
    old_mole_fraction=None,
    inplace=False,
    ignore_warnings=False,
    force=False,
    verbose=True,
):
    """Update spectrum with new molar fraction Convoluted values (with slit)
    are dropped in the process.

    Parameters
    ----------

    spec: Spectrum

    new_mole_fraction: float
        new mole fraction

    old_mole_fraction: float, or None
        if None, current mole fraction (conditions['mole_fraction']) is used


    Other Parameters
    ----------------

    inplace: boolean
        if ``True``, modifies the Spectrum object directly. Else, returns
        a copy. Default ``False``.

    force: boolean
        if False, won't allow rescaling to 0 (not to loose information).
        Default ``False``

    Returns
    -------

    s_rescaled: Spectrum
        a rescaled Spectrum.
        if ``inplace=True``, then ``s`` has been rescaled already and
        ``s_rescaled`` is ``s``

    Notes
    -----

    Implementation:

        similar to rescale_path_length() but we have to scale abscoeff & emisscoeff
        Note that this is valid only for small changes in mole fractions. Then,
        the change in line broadening becomes significant
    """

    if not inplace:
        spec = spec.copy()

    # Check inputs
    # ---------
    # ... get old mole fraction, use existing one if not given
    if old_mole_fraction is not None:
        try:
            if (
                spec.conditions["mole_fraction"] != old_mole_fraction
                and not ignore_warnings
            ):
                warn(
                    "mole_fraction ({0}) doesnt match value given in conditions ({1})".format(
                        old_mole_fraction, spec.conditions["mole_fraction"]
                    )
                )
        except KeyError:  # mole fraction not defined
            pass

    else:
        try:
            old_mole_fraction = spec.conditions["mole_fraction"]
        except KeyError:
            raise KeyError(
                "mole_fraction has to be defined in conditions (or use"
                + " `from_mole_fraction`)"
            )

    # Add warning is large mole fraction rescale
    if np.abs(new_mole_fraction - old_mole_fraction) > 0.3:
        warn(
            "Large rescaling from {0} to {1}. ".format(
                old_mole_fraction, new_mole_fraction
            )
            + "There may be significant changes in pressure broadening coefficients."
            + "You should calculate a new spectrum instead."
        )

    if new_mole_fraction < 0 and not force:
        raise ValueError("mole_fraction cannot be negative")
    if new_mole_fraction == 0 and not force:
        raise ValueError(
            "Rescaling to 0 will loose information. Choose force " "= True"
        )
    if new_mole_fraction > 1 and not force:
        warn("rescaling to mole fraction > 1: {0}".format(new_mole_fraction))

    for q in ["transmittance", "radiance"]:
        qns = q + "_noslit"
        qties = spec.get_vars()
        if q in qties and qns not in qties and not force:
            raise KeyError(
                "Cant rescale {0} if {1} not stored.".format(q, qns)
                + "(you need to rescale before applying the slit again) "
                + " Use force=True to rescale anyway, but {0}".format(q)
                + " will be deleted"
            )

    # Get path length
    if "path_length" in list(spec.conditions.keys()):
        path_length = spec.conditions["path_length"]
        true_path_length = True
    else:
        path_length = 1
        true_path_length = False

    # Rescale
    assume_equilibrium = spec.conditions.get("thermal_equilibrium", False)
    _recalculate(
        spec,
        "same",
        path_length,
        path_length,
        new_mole_fraction,
        old_mole_fraction,
        true_path_length=true_path_length,
        assume_equilibrium=assume_equilibrium,
        verbose=verbose,
    )

    # Update conditions
    spec.conditions["mole_fraction"] = new_mole_fraction

    return spec


if __name__ == "__main__":

    from radis.test.spectrum.test_rescale import _run_all_tests

    print(("Test rescale.py: ", _run_all_tests(verbose=True)))



# -*- coding: utf-8 -*-
"""Misc. and support functions
"""


from .arrays import (
    array_allclose,
    autoturn,
    bining,
    calc_diff,
    centered_diff,
    count_nans,
    evenly_distributed,
    find_first,
    find_nearest,
    is_sorted,
    is_sorted_backward,
    logspace,
    nantrapz,
    norm,
    norm_on,
    scale_to,
)
from .basics import (
    compare_dict,
    compare_lists,
    compare_paths,
    exec_file,
    is_float,
    key_max_val,
    list_if_float,
    make_folders,
    merge_lists,
    partition,
    remove_duplicates,
)
from .config import getDatabankEntries, getDatabankList
from .curve import (
    curve_add,
    curve_distance,
    curve_divide,
    curve_multiply,
    curve_substract,
)
from .debug import export
from .progress_bar import ProgressBar
from .signal import resample, resample_even
from .utils import DatabankNotFound, NotInstalled, getProjectRoot



# -*- coding: utf-8 -*-
"""Created on Thu Sep 14 13:44:35 2017.

@author: erwan

Summary
-------

An experimental module to merge precalculated vibrational bands in post-processing,
and recompute a new spectrum at a new temperature or with overpopulations, without
having to recalculate the broadening of each line

.. warning::

    Order of magnitude faster, but only valid under optically thin conditions as the
    rescaling of absorption doesnt scale induced emission properly

----------
"""

from time import time
from warnings import warn

import numpy as np
import pandas as pd
from numpy import exp

try:  # Proper import
    from .labels import vib_lvl_name_cdsd_pc, vib_lvl_name_cdsd_pcN
except ImportError:  # if ran from here
    from radis.lbl.labels import vib_lvl_name_cdsd_pc, vib_lvl_name_cdsd_pcN
from radis.misc.basics import is_float
from radis.phys.constants import hc_k
from radis.spectrum.rescale import (
    rescale_abscoeff,
    rescale_absorbance,
    rescale_emisscoeff,
    rescale_radiance_noslit,
    rescale_transmittance_noslit,
)
from radis.spectrum.spectrum import Spectrum

# keys that may not be equals for different bands
_IGNORE_KEYS = ["band", "band_htrn", "viblvl_u", "viblvl_l"]


class LevelsList(object):
    """A class to generate a Spectrum from a list of precalculated bands at a
    given reference temperature.

    .. warning::

        only valid under optically thin conditions!!

    See Also
    --------
    """

    # hardcode attribute names, but can save a lot of memory
    __slots__ = [
        "Tvib_ref",
        "Trot_ref",
        "vib_distribution_ref",
        "mole_fraction_ref",
        "path_length_ref",
        "sorted_bands",
        "cum_weight",
        "E_bands",
        "bands_ref",
        "verbose",
        "bands",
        "parfunc",
        "levelsfmt",
        "lvl_index",
        "vib_levels",
        "copy_lines",
        "Qref",
        "Qvib_ref",
    ]

    def __init__(
        self, parfunc, bands, levelsfmt, sortby="Ei", copy_lines=False, verbose=True
    ):
        """
        Parameters
        ----------
        bands: dict of bands
            bands are Spectrum objects calculated at equilibrium or non-equilibrim.
            Tgas, or (Tvib, Trot) must be given and the same in all bands conditions.

        """

        # Check inputs
        for br, s in bands.items():
            if not isinstance(s, Spectrum):
                raise ValueError(
                    "`bands` must be a list of Spectrum objects. "
                    + "Got {0}".format(type(s))
                )
            if s.lines is None:
                raise ValueError(
                    "To be used in LevelsList spectra must have been calculated with 'export_lines=True'"
                )

        # Assert all conditions are the same
        bands_list = list(bands.values())
        cond_ref = bands_list[0].conditions
        for s in bands_list[1:]:  # type(s): Spectrum
            if list(s.conditions.keys()) != list(cond_ref.keys()):
                raise ValueError("Input conditions must be the same for all bands")
            for k in cond_ref.keys():
                if k in _IGNORE_KEYS:  # don't compare these entries
                    continue
                if s.conditions[k] != cond_ref[k]:
                    raise ValueError(
                        "Input conditions must be the same for all bands"
                        + ". Got a difference for entry {0}: {1} vs {2}".format(
                            k, s.conditions[k], cond_ref[k]
                        )
                    )

        if "Tvib" in cond_ref:
            Tvib_ref = cond_ref["Tvib"]
        else:
            Tvib_ref = cond_ref["Tgas"]
        if "Trot" in cond_ref:
            Trot_ref = cond_ref["Trot"]
        else:
            Trot_ref = cond_ref["Tgas"]
        vib_distribution_ref = cond_ref["vib_distribution"]
        if "overpopulation" in cond_ref:
            raise NotImplementedError

        # stores input conditions
        self.Tvib_ref = Tvib_ref
        self.Trot_ref = Trot_ref
        self.mole_fraction_ref = cond_ref["mole_fraction"]
        self.path_length_ref = cond_ref["path_length"]
        self.vib_distribution_ref = vib_distribution_ref

        # stores computation parameters
        # TODO: store it in parfunc instead (and make parfunc an EnergyDatabase)
        self.levelsfmt = levelsfmt

        # %% Misc
        self.verbose = verbose
        self.copy_lines = copy_lines

        # %% Set up energies  ( use a partition function object for the moment)

        self.parfunc = parfunc
        self._init_levels(sortby)
        self._connect_bands_to_levels(bands)

        Qvib = self._calc_vib_populations(
            Tvib=self.Tvib_ref, vib_distribution=self.vib_distribution_ref
        )

        self.Qref = parfunc.at_noneq(
            Tvib_ref, Trot_ref, vib_distribution=vib_distribution_ref
        )  # TODO: add overpopulations in partition function
        self.Qvib_ref = Qvib

        self.vib_levels["nvib_ref"] = self.vib_levels["nvib"]
        del self.vib_levels["nvib"]

        # %% Set up bands
        self._init_bands_ref(bands, sortby)

    def _init_bands_ref(self, bands, sortby):

        # %% Parse bands

        #        # Get band names   (if bands was a list)
        #        band_dict = {}
        #        for bd in bands:
        #            try:
        #                band_name = set(bd.lines['band'])
        #            except:
        #                raise ValueError('Bands must have been created with a `band` attribute'+\
        #                                 ' in lines')
        #            assert len(band_name) == 0
        #            band_dict[band_name[0]] = bd
        #        self.bands = band_dict
        #        self.bands0 = {br: bands[br].copy(copy_lines=False) for br in bands}  # make a copy

        # Get energy of upper level of band
        E_bands = {}
        for br, band in bands.items():  # type(band): Spectrum
            if br == "others":
                continue
            E_bands[br] = np.mean(list(set(band.lines["Evibu"])))
            # Todo: should be all equals ?
            assert len(set(band.lines["Evibu"])) == 1

        # Calculate band weight
        weight = {}
        for br, band in bands.items():  # type(band): Spectrum
            weight[br] = band.lines[sortby].sum()
        tot = np.sum([np.array(w) for w in weight.values()])
        for br, band in bands.items():
            weight[br] /= tot
        sorted_bands = sorted(bands, key=lambda br: weight[br])[::-1]
        cum_weight = np.array([weight[br] for br in sorted_bands])
        cum_weight = np.cumsum(cum_weight)

        # Store
        self.sorted_bands = sorted_bands
        self.cum_weight = cum_weight
        self.E_bands = E_bands

        #        # Delete lines (makes copies faster)
        #        for br in self.bands0:
        #            del self.bands0[br].lines

        # Initialize bands
        self.bands_ref = {
            br: band.copy(copy_lines=self.copy_lines) for br, band in bands.items()
        }

    def _init_levels(self, sortby):
        """
        Notes
        -----

        vib level in (p,c,j,N) notation is ambiguous. Here we use (p,c) based
        on energy difference and perturbation rules
        """

        df = self.parfunc.df
        levelsfmt = self.levelsfmt

        if levelsfmt == "cdsd-pc":
            vib_index = ["p", "c"]

            def vib_lvl_name_cdsd(p, c, N):
                return vib_lvl_name_cdsd_pc(p, c)

        elif levelsfmt == "cdsd-pcN":
            vib_index = ["p", "c", "N"]
            vib_lvl_name_cdsd = vib_lvl_name_cdsd_pcN

        else:
            raise NotImplementedError(levelsfmt)

        vib_levels = df.drop_duplicates(subset=vib_index)  # 192 ms ± 1.13 ms

        # Add levels
        def add_viblvl(row):
            """"""
            #    row['vib_lvl'] = '({p},{c},{N})'.format(**dict([(k,int(row[k])) for k in ['p', 'c', 'N']]))
            #            row['viblvl'] = '({p},{c})'.format(**dict([(k,int(row[k])) for k in ['p', 'c']]))
            row["viblvl"] = vib_lvl_name_cdsd(row.p, row.c, row.N)
            return row

        t0 = time()
        vib_levels = vib_levels.apply(add_viblvl, axis=1)
        print("Added viblvl in {0:.2f}s".format(time() - t0))

        # vib_levels = vib_levels.set_index(['p', 'c', 'N'])
        vib_levels = vib_levels.set_index(["viblvl"])

        # %% Generate index of  levels

        lvl_index = dict().fromkeys(list(vib_levels.index))
        for (
            k
        ) in (
            lvl_index
        ):  # Note: do not fill directly in frmokeys or you'll get the same shared list
            lvl_index[k] = {"bands_where_low": [], "bands_where_up": []}

        self.lvl_index = lvl_index
        self.vib_levels = vib_levels

    def _connect_bands_to_levels(self, s_bands):

        index = self.lvl_index

        for s in s_bands.values():
            if s.name == "others":
                continue

            viblvl_u = s.lines["viblvl_u"].iloc[0]
            viblvl_l = s.lines["viblvl_l"].iloc[0]

            index[viblvl_u]["bands_where_up"].append(s)
            index[viblvl_l]["bands_where_low"].append(s)

    def plot_vib_populations(self, nfig=None, **kwargs):
        """Plot current distribution of vibrational levels.

        By constructions populations are shown as divided by the state degeneracy,
        i.e,            g =   gv   *   (2J+1)   *    gi    *     gs

        Parameters
        ----------

        nfig: str, int
            name of Figure to plot on

        kwargs: **dict
            arguments are forwarded to plot()
        """

        import matplotlib.pyplot as plt

        vib_levels = self.vib_levels

        plt.figure(num=nfig)
        plt.plot(vib_levels.E, vib_levels.nvib, "ok", **kwargs)
        plt.xlabel("Energy (cm-1)")
        plt.ylabel("Population")
        plt.yscale("log")
        plt.xlim(xmin=0)

        return

    def _calc_vib_populations(
        self, Tvib, vib_distribution="boltzmann", overpopulation=None
    ):
        """Calculate vibrational populations from Tref to new Tvib and store
        results in vib_levels dataframe This does not modify the spectra yet!

        By constructions populations are calculated divided by the state degeneracy,
        i.e,            g =   gv   *   (2J+1)   *    gi    *     gs
        This means we should only use ratios for rescaling

        The information on state degeneracy and isotopic abundance is already
        included in the linestrength / emission integral, hence in the pre-calculated
        emisscoeff / abscoeff
        """

        vib_levels = self.vib_levels
        if overpopulation is None:
            overpopulation = {}

        if is_float(Tvib):

            # Get new population
            E_vib = vib_levels["Evib"]
            g = 1  # explicitely calculate populations divided by degeneracy
            # this means we should only use ratios for rescaling
            if vib_distribution == "boltzmann":
                nvibQvib = g * exp(-hc_k * E_vib / Tvib)
            else:
                raise NotImplementedError(
                    "vib_distribution: {0}".format(vib_distribution)
                )

        else:
            Tvib1, Tvib2, Tvib3 = Tvib

            # Get new population
            E_vib1 = vib_levels["Evib1"]
            E_vib2 = vib_levels["Evib2"]
            E_vib3 = vib_levels["Evib3"]
            g = 1  # explicitely calculate populations divided by degeneracy
            # this means we should only use ratios for rescaling
            if vib_distribution == "boltzmann":
                nvibQvib = (
                    g
                    * exp(-hc_k * E_vib1 / Tvib1)
                    * exp(-hc_k * E_vib2 / Tvib2)
                    * exp(-hc_k * E_vib3 / Tvib3)
                )
            else:
                raise NotImplementedError(
                    "vib_distribution: {0}".format(vib_distribution)
                )

        # Add overpopulation
        if overpopulation != {}:
            for k, ov in overpopulation.items():
                nvibQvib.loc[k] *= ov
                # TODO: test
            warn(
                "NotImplemented: partition function overpopulation correction not tested"
            )

        # Normalize with partition function
        Qvib = nvibQvib.sum()
        nvib = nvibQvib / Qvib

        # update dataframe
        vib_levels["nvib"] = nvib
        vib_levels["Qvib"] = Qvib

        return Qvib

    #    def set_cutoff(self, cutoff=1e-3):
    #        ''' Discard most of the bands and keep error below cutoff
    #
    #        Warning: this cannot be undone!'''
    #
    #        cum_weight = self.cum_weight
    #        sorted_bands = self.sorted_bands
    #
    #        cutoff_i = np.argmax((1-cum_weight)<cutoff)
    #        first_bands = sorted_bands[:cutoff_i+1]
    #
    #        # Update bands
    #        new_bands = {br: self.bands_ref[br] for br in first_bands}
    #
    #        if self.verbose:
    #            print(('Set cutoff. Discarded {0}/{1} bands. Estimated error: {2:.2f}%'.format(
    #                len(self.bands_ref)-len(new_bands), len(self.bands_ref),
    #                (1-cum_weight[cutoff_i])*100)))
    #
    #        self.bands_ref = new_bands

    def eq_spectrum(
        self,
        Tgas,
        overpopulation=None,
        mole_fraction=None,
        path_length=None,
        save_rescaled_bands=False,
    ):
        """See :py:meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`

        .. warning::

            only valid under optically thin conditions!!

        Parameters
        ----------

        ... same as usually. If None, then the reference value (used to
        calculate bands) is used

        Other Parameters
        ----------------

        save_rescaled_bands: boolean
            save updated bands. Take some time as it requires rescaling all
            bands individually (which is only done on the MergedSlabs usually)
            Default ``False``
        """
        s = self.non_eq_spectrum(
            Tgas,
            Tgas,
            overpopulation=overpopulation,
            mole_fraction=mole_fraction,
            path_length=path_length,
            save_rescaled_bands=save_rescaled_bands,
        )
        del s.conditions["Tvib"]
        del s.conditions["Trot"]
        s.conditions["Tgas"] = Tgas

        return s

    def non_eq_spectrum(
        self,
        Tvib=None,
        Trot=None,
        Ttrans=None,
        vib_distribution="boltzmann",
        overpopulation=None,
        mole_fraction=None,
        path_length=None,
        save_rescaled_bands=False,
    ):
        """See :py:meth:`~radis.lbl.factory.SpectrumFactory.non_eq_spectrum`

        .. warning::

            only valid under optically thin conditions!!

        Parameters
        ----------

        ... same as usually. If None, then the reference value (used to
        calculate bands) is used

        Other Parameters
        ----------------

        save_rescaled_bands: boolean
            save updated bands. Take some time as it requires rescaling all
            bands individually (which is only done on the MergedSlabs usually)
            Default ``False``

        Notes
        -----

        Implementation:

        Generation of a new spectrum is done by recombination of the precalculated
        bands with
        """

        from radis.los import MergeSlabs

        # Restart from a copy each time  (else reference bands are modified by rescaling
        # and we may loose information if rescaling to 0 for instance)
        bands_ref = self.bands_ref
        bands = {br: bands_ref[br].copy(copy_lines=self.copy_lines) for br in bands_ref}

        # Initialize inputs
        if Trot is None:
            Trot = self.Trot_ref
        if Tvib is None:
            Tvib = self.Tvib_ref
        if Trot != self.Trot_ref:
            raise ValueError(
                "Trot {0} doesnt match the reference Trot: {1}".format(
                    Trot, self.Trot_ref
                )
            )
        if mole_fraction is None:
            mole_fraction = self.mole_fraction_ref
        if path_length is None:
            path_length = self.path_length_ref
        if overpopulation is None:
            overpopulation = {}

        #        # Fill missing overpopulation with 1
        #        if overpopulation is None:
        #            overpopulation = dict.fromkeys(list(bands.keys()), 1)
        #        else:
        #            for br in bands.keys():
        #                if not br in overpopulation:
        #                    overpopulation[br] = 1

        #        Tvib_ref = self.Tvib_ref
        #        pop_correction = dict.fromkeys(list(bands.keys()), 1)
        #        if Tvib != Tvib_ref:
        #            raise NotImplementedError
        #            E_bands = self.E_bands
        #            # Correct for vibrational temperature
        #            for br in bands:
        #                pop_correction[br] = exp(-E_bands[br]*hc_k/Tvib)/exp(-E_bands[br]*hc_k/Tvib_ref)

        # Recalculate populations from reference everytime
        vib_levels = self.vib_levels

        # Recalculate partition function
        Qvib = self._calc_vib_populations(
            Tvib=Tvib, vib_distribution=vib_distribution, overpopulation=overpopulation
        )
        # TODO: note: Qvib doesnt take gvib into account. Wrong Qvib but correct populations
        # if rescaled with ratio???
        # ... warning: update_populations set to False here not updated (because we dont
        # ... care much about all levels here. only the one from vib_levels matter)
        if overpopulation is not None:
            Q, _, _ = self.parfunc.at_noneq(
                Tvib,
                Trot,
                overpopulation=overpopulation,
                returnQvibQrot=True,
                update_populations=False,
            )
        else:
            Q = self.parfunc.at_noneq(Tvib, Trot, update_populations=False)

        Qref = self.Qref
        Qvib_ref = self.Qvib_ref

        for br, band in bands.items():  # type(band): Spectrum
            if br == "others":
                continue
            viblvl_u = band.conditions["viblvl_u"]
            viblvl_l = band.conditions["viblvl_l"]
            nu_vib = vib_levels.loc[viblvl_u, "nvib"]
            nu_vib_old = vib_levels.loc[viblvl_u, "nvib_ref"]
            nl_vib = vib_levels.loc[viblvl_l, "nvib"]
            nl_vib_old = vib_levels.loc[viblvl_l, "nvib_ref"]

            # with new, old = (2, 1):
            # n2/n1 = exp(-Evib2/kT2)/exp(-Evib1/kT1) * Qvib1/Qvib2
            #       = n2vib / n1vib * Q1/Q2 * Q2vib / Q1vib
            if vib_distribution == "boltzmann":
                corfactor_u = nu_vib / nu_vib_old * Qref / Q * Qvib / Qvib_ref
                corfactor_l = nl_vib / nl_vib_old * Qref / Q * Qvib / Qvib_ref
            else:
                raise NotImplementedError(
                    "vib_distribution: {0}".format(vib_distribution)
                )

            rescale_updown_levels(band, corfactor_u, 1, corfactor_l, 1)

            # Update lines
            if self.copy_lines:
                band.lines["nu"] *= corfactor_u
                band.lines["nl"] *= corfactor_l
                band.lines["Qvib"] = Qvib
                band.lines["nu_vib"] *= corfactor_u
                band.lines["nl_vib"] *= corfactor_l
                band.lines["Ei"] *= np.nan
                band.lines["S"] = np.nan

        # Get total spectrum
        s = MergeSlabs(*list(bands.values()))

        # populations
        s.populations = pd.DataFrame(vib_levels[["nvib", "Evib"]])

        # rebuild lines
        if self.copy_lines:
            s.lines = pd.concat([band.lines for band in bands.values()])

        # Update total mole fraction if it changed
        mole_fraction_ref = self.mole_fraction_ref
        if mole_fraction != mole_fraction_ref:
            s.rescale_mole_fraction(
                mole_fraction,
                mole_fraction_ref,
                ignore_warnings=True,  # mole_fraction is 'N/A'
                force=True,
            )

        path_length_ref = self.path_length_ref
        if path_length != path_length_ref:
            s.rescale_path_length(path_length, path_length_ref, force=True)

        # Add parameters in conditions:
        s.conditions["overpopulation"] = overpopulation
        s.conditions["mole_fraction"] = mole_fraction  # was 'N/A' after Merge
        # because it's different for all bands
        s.conditions["path_length"] = path_length
        s.conditions["Tvib"] = Tvib
        s.conditions["Trot"] = Trot

        if save_rescaled_bands:
            for br, band in bands.items():  # type(band): Spectrum
                if mole_fraction != mole_fraction_ref:
                    band.rescale_mole_fraction(
                        mole_fraction,
                        mole_fraction_ref,
                        ignore_warnings=True,  # mole_fraction is 'N/A'
                        force=True,
                    )
                if path_length != path_length_ref:
                    band.rescale_path_length(path_length, path_length_ref, force=True)
            self.bands = bands

        return s


# %% Overpopulation functions


def rescale_updown_levels(
    spec, new_nu, old_nu, new_nl, old_nl, ignore_warnings=False, force=False
):
    """Update spectrum with new molar fraction for upper and lower levels.

    Convoluted values (with slit) are dropped in the process.

    Rescales with a ratio new_nu/old_nu.

    This is only valid for emission quantities, under optically thin conditions,
    as rescaling doesnt correct for induced emission.

    .. warning::
        experimental feature

    Parameters
    ----------

    new_nu: float
        new upper state mole fraction

    old_nu: float
        current upper state mole fraction

    new_nl: float
        new lower state mole fraction

    old_nl: float
        current lower state mole fraction

    Other Parameters
    ----------------

    force: boolean
        if False, won't allow rescaling to 0 (not to loose information).
        Default ``False``

    Notes
    -----

    Implementation:

    similar to rescale_path_length() but we have to scale abscoeff & emisscoeff
    Note that this is valid only for small changes in mole fractions. Then,
    the change in line broadening becomes significant and the whole band
    should be recomputed

    IMPORTANT: if editing make sure you use the proper nu and nl. In particular
    when infering emission quantities from absorption quantities this may
    ends up in error in overpopulation rescaling.
    """
    # TODO: Add warning when too large rescaling

    # Check inputs
    # ---------
    #        if old_mole_fraction is not None:
    #            try:
    #                if spec.conditions['mole_fraction'] != old_mole_fraction and not ignore_warnings:
    #                    warn('mole_fraction ({0}) doesnt match value given in conditions ({1})'.format(
    #                            old_mole_fraction, spec.conditions['mole_fraction']))
    #            except KeyError: # mole fraction not defined
    #                pass

    #        else:
    #            try:
    #                old_mole_fraction = spec.conditions['mole_fraction']
    #            except KeyError:
    #                raise KeyError('mole_fraction has to be defined in conditions (or use'+\
    #                                ' `from_mole_fraction`)'),

    if (new_nu < 0 or new_nl < 0) and not force:
        raise ValueError("mole_fraction cannot be negative")
    if (new_nu == 0 or new_nl == 0) and not force:
        raise ValueError(
            "Rescaling to 0 will loose information. Choose force " "= True"
        )

    for q in ["transmittance", "radiance"]:
        qns = q + "_noslit"
        qties = spec.get_vars()
        if q in qties and qns not in qties and not force:
            raise KeyError(
                "Cant rescale {0} if {1} not stored".format(q, qns)
                + " Use force=True to rescale anyway. {0}".format(q)
                + " will be deleted"
            )

    # Get path length
    if "path_length" in list(spec.conditions.keys()):
        path_length = spec.conditions["path_length"]
        true_path_length = True
    else:
        path_length = 1
        true_path_length = False

    # %%  Rescale

    optically_thin = spec.is_optically_thin()

    # Choose which values to recompute
    # ----------
    initial = spec.get_vars()  # quantities initialy in spectrum
    recompute = list(initial)  # quantities to recompute
    rescaled = {}  # quantities rescaled

    if "radiance_noslit" in initial and not optically_thin:
        recompute.append("emisscoeff")
    if (
        "absorbance" in initial
        or "transmittance_noslit" in initial
        or "radiance_noslit" in initial
        and not optically_thin
    ):
        recompute.append("abscoeff")
    extra = []
    recompute = set(recompute)  # remove duplicates

    # Get units
    units = spec.units.copy()

    # Recompute!
    # ----------
    waveunit = spec.get_waveunit()  # keep all quantities in same waveunit

    if "abscoeff" in recompute:
        rescaled, units = rescale_abscoeff(
            spec,
            rescaled,
            initial,  # TODO: remove rescaled = ... Dict is mutable
            old_nl,
            new_nl,
            path_length,
            waveunit,
            units,
            extra,
            true_path_length,
        )

    if "emisscoeff" in recompute:
        rescaled, units = rescale_emisscoeff(
            spec,
            rescaled,
            initial,
            old_nu,
            new_nu,
            path_length,
            optically_thin,
            waveunit,
            units,
            extra,
            true_path_length,
        )

    if "absorbance" in recompute:
        rescaled, units = rescale_absorbance(
            spec,
            rescaled,
            initial,
            old_nl,
            new_nl,
            path_length,
            path_length,
            waveunit,
            units,
            extra,
            true_path_length,
        )

    if "transmittance_noslit" in recompute:
        rescaled, units = rescale_transmittance_noslit(
            spec,
            rescaled,
            initial,
            old_nl,
            new_nl,
            path_length,
            path_length,
            waveunit,
            units,
            extra,
            true_path_length,
        )

    if "radiance_noslit" in recompute:
        rescaled, units = rescale_radiance_noslit(
            spec,
            rescaled,
            initial,
            old_nu,
            new_nu,
            path_length,
            path_length,
            optically_thin,
            waveunit,
            units,
            extra,
            true_path_length,
        )

    # Save (only) the ones that were in the spectrum initially
    for q in rescaled:
        if q in initial:
            spec._q[q] = rescaled[q]

    # Update units
    for k, u in units.items():
        spec.units[k] = u

    # Drop convoluted values
    for q in ["transmittance", "radiance"]:
        if q in list(spec._q.keys()):
            del spec._q[q]

    # Reapply slit if possible
    if (
        "slit_function" in spec.conditions
        and "slit_unit" in spec.conditions
        and "norm_by" in spec.conditions
    ):
        slit_function = spec.conditions["slit_function"]
        slit_unit = spec.conditions["slit_unit"]
        norm_by = spec.conditions["norm_by"]
        try:
            shape = spec.conditions["shape"]
        except KeyError:
            shape = "triangular"
        spec.apply_slit(
            slit_function=slit_function, unit=slit_unit, shape=shape, norm_by=norm_by
        )

    # Update conditions
    spec.conditions["mole_fraction"] = new_nl if new_nl == new_nu else "N/A"


# def _test(verbose=True, plot=False, *args, **kwargs):
#    """ Generate scalable LevelList at 1500K. Rescale at 2000K and compare with
#    spectrum directly calculated at 2000K
#    """
#
#    from radis.lbl import SpectrumFactory
#
#    iso = 1
#    sf = SpectrumFactory(
#        wavelength_min=4170,
#        wavelength_max=4200,
#        mole_fraction=1,
#        path_length=0.05,
#        cutoff=1e-25,
#        #                     isotope=[1,2],
#        isotope=iso,
#        db_use_cached=True,
#        wstep=0.01,
#        broadening_max_width=10,
#        medium="air",
#        verbose=verbose,
#    )
#
#    sf.load_databank("CDSD")
#
#    parfunc = sf.parsum_calc["CO2"][iso]["X"]
#
#    # %% Fill levels for all bands
#
#    Tref = 1500
#    s_bands = sf.non_eq_bands(Tvib=Tref, Trot=Tref)
#    lvlist = LevelsList(parfunc, s_bands, sf.params.levelsfmt)
#
#    # %% Test
#    Tvib = 2000
#    s_resc = lvlist.non_eq_spectrum(Tvib=Tvib, Trot=Tref)
#    s0 = sf.non_eq_spectrum(Tvib, Tref)
#
#    return s0.compare_with(s_resc, spectra_only=True, plot=plot)


if __name__ == "__main__":

    from radis.test.lbl.test_overp import run_testcases

    print("test_overp.py:", run_testcases(plot=True))



# -*- coding: utf-8 -*-
"""
Determine gas mixture composition under chemical equilibrium using CANTERA.

-------------------------------------------------------------------------------

"""


def get_eq_mole_fraction(initial_mixture, T_K, p_Pa):
    """Calculates chemical equilibrium mole fraction at temperature T, using
    the CANTERA :py:meth:`~cantera.ThermoPhase.equilibrate` function.

    The calculation uses the default GRI3.0 mechanism, which was
    designed to model natural gas combustion, including NO formation
    and reburn chemistry. See `GRI 3.0 <combustion.berkeley.edu/gri-mech/version30/text30.html>`__.

    When using, cite the [CANTERA]_ package.

    Parameters
    ----------

    initial_mixture: str
        Gas composition. Example::

             'N2:0.79, O2:0.21, CO2:363e-6'

        Or::

             'CO2:1'

    T_K: float (K)
        temperature (Kelvin) to calculate equilibrium

    P_Pa: float (Pa)
        temperature (Pascal) to calculate equilibrium

    Examples
    --------

    Calculate equilibrium mixture of CO2 at 2000 K, 1 atm::

        get_eq_mole_fraction('CO2:1', 2000, 101325)

        >>> {'C': 1.7833953335281855e-19,
            'CO': 0.01495998583472384,
            'CO2': 0.9775311634424326,
            'O': 5.7715610124613225e-05,
            'O2': 0.007451135112719029}

    References
    ----------

    [CANTERA]_
    """

    try:
        import cantera as ct
    except ImportError as err:
        raise ImportError(
            "Cantera is needed to calculate equilibrium mole fractions"
            + ". Install with  `pip install cantera` or (better) `conda install -c cantera cantera`",
        ) from err

    # %% Init Cantera
    g = ct.Solution("gri30.xml")
    g.TPX = T_K, p_Pa, initial_mixture

    # Minimize Gibbs:
    g.equilibrate("TP")

    # Returns
    return g.mole_fraction_dict()



"""Install file for RADIS.

Typical install procedure, plus:

- auto-convert README.rst to long_description, removing some sphinx-only syntax
so it can be rendered by PyPi
- read version number from __version__.txt
- some dependencies should be installed manually (typically: all packages with
compiled components such as numpy, pandas, etc.)


Examples
--------

Install (normal, use-only)::

    python setup.py install

Or (create an alias, so you can still edit)::

    python setup.py develop

Notes
-----

For developers:

when creating a new version, just update the __version__.txt file

to register it on Pypi see register.py::

    python register.py
"""
import io
import re
import sys
from os.path import abspath, dirname, exists, join

from numpy import get_include
from setuptools import Extension, find_packages, setup

# Build description from README (PyPi compatible)
# -----------------------------------------------

# Utils to format RST
def yield_sphinx_only_markup(lines):
    """Cleans-up Sphinx-only constructs (ie from README.rst), so that *PyPi*
    can format it properly.

    To check for remaining errors, install ``sphinx`` and run::

        python setup.py --long-description | sed -file 'this_file.sed' | rst2html.py  --halt=warning

    :param file_inp:     a `filename` or ``sys.stdin``?
    :param file_out:     a `filename` or ``sys.stdout`?`

    References
    ----------

    https://stackoverflow.com/questions/16367770/my-rst-readme-is-not-formatted-on-pypi-python-org

    Notes
    -----

    Check output with::

        python setup.py --long-description | rst2html.py > output.html
    """
    substs = [
        ## Selected Sphinx-only Roles.
        #
        (r":abbr:`([^`]+)`", r"\1"),
        (r":ref:`([^`]+)`", r"`\1`_"),
        (r":term:`([^`]+)`", r"**\1**"),
        (r":dfn:`([^`]+)`", r"**\1**"),
        (r":(samp|guilabel|menuselection):`([^`]+)`", r"``\2``"),
        ## Sphinx-only roles:
        #        :foo:`bar`   --> foo(``bar``)
        #        :a:foo:`bar` XXX afoo(``bar``)
        #
        # (r'(:(\w+))?:(\w+):`([^`]*)`', r'\2\3(``\4``)'),
        (r":(\w+):`([^`]*)`", r"\1(``\2``)"),
        ## Sphinx-only Directives.
        #
        (r"\.\. doctest", r"code-block"),
        (r"\.\. plot::", r".. "),
        (r"\.\. seealso", r"info"),
        (r"\.\. glossary", r"rubric"),
        (r"\.\. figure::", r".. "),
        ## Other
        #
        (r"\|version\|", r"x.x.x"),
        ## added to make RADIS docs Pypi compatible
        #        (r'\.\. image::',          r'.. '),
        #        (r'\.\. |CO2| replace:: CO\ :sub:`2`',          r'.. '),
    ]

    regex_subs = [(re.compile(regex, re.IGNORECASE), sub) for (regex, sub) in substs]

    def clean_line(line):
        try:
            for (regex, sub) in regex_subs:
                line = regex.sub(sub, line)
        except Exception as ex:
            print(("ERROR: %s, (line(%s)" % (regex, sub)))
            raise ex

        return line

    for line in lines:
        yield clean_line(line)


# (note: README.rst has been converted to README.md by register.py, and cleaned afterwards )
description = "A fast line-by-line code for high-resolution infrared molecular spectra"
readme_path = join(abspath(dirname(__file__)), "README.md")
if not exists(readme_path):
    long_description = description
else:
    with io.open(readme_path, encoding="utf-8") as f:
        long_description = f.read()

# Read version number from file
with open(join(dirname(__file__), "radis", "__version__.txt")) as version_file:
    __version__ = version_file.read().strip()


# %% Cython extensions:
# Ensures RADIS still builds if without Cython


class BuildFailed(Exception):
    pass


from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError


def show_message(*lines):
    """ Note : will only happen if user installs with `pip install -v` """
    print("=" * 74, file=sys.stderr)
    for line in lines:
        print(line, file=sys.stderr)
    print("=" * 74, file=sys.stderr)


def get_ext_modules(with_binaries):
    """
    Parameter
    ---------
    with_binaries: bool
        if False, do not try to build Cython extensions
    """
    # Based on code from https://github.com/pallets/markupsafe/blob/main/setup.py

    print(sys.version)
    ext_modules = []
    cmdclass = {}

    if not with_binaries:
        # skip building extensions
        return {"cmdclass": cmdclass, "ext_modules": ext_modules}

    # TO-DO: set language level

    try:
        import cython
    except (ModuleNotFoundError) as err:
        raise BuildFailed(
            "Cython not found : Skipping all Cython extensions...!"
        ) from err

    print("Cython " + cython.__version__)

    from Cython.Distutils import build_ext

    class build_ext_subclass(build_ext):
        def build_extensions(self):
            c = self.compiler.compiler_type
            copt = {
                "msvc": ["/openmp", "/Ox", "/fp:fast", "/favor:INTEL64"],
                "mingw32": ["-fopenmp", "-O3", "-ffast-math", "-march=native"],
            }

            lopt = {"mingw32": ["-fopenmp"]}

            print("Compiling with " + c + "...")
            try:
                for e in self.extensions:
                    e.extra_compile_args = copt[c]
            except (KeyError):
                pass
            try:
                for e in self.extensions:
                    e.extra_link_args = lopt[c]
            except (KeyError):
                pass
            try:
                build_ext.build_extensions(self)
            except (CCompilerError, DistutilsExecError, DistutilsPlatformError) as err:
                raise BuildFailed() from err

    ext_modules.append(
        Extension(
            "radis_cython_extensions",
            sources=["./radis/cython/radis_cython_extensions.pyx"],
            include_dirs=[get_include()],
            language="c++",
            extra_link_args=[],
        )
    )

    cmdclass["build_ext"] = build_ext_subclass

    return {"cmdclass": cmdclass, "ext_modules": ext_modules}


#%% Main install routine
def run_setup(with_binary):
    setup(
        name="radis",
        version=__version__,
        description=description,
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/radis/radis",
        author="Erwan Pannier",
        author_email="erwan.pannier@gmail.com",
        license="GNU Lesser General Public License v3 (LGPLv3)",
        keywords=[
            "spectrum",
            "infrared",
            "spectra",
            "radiation",
            "nonequilibrium",
            "spectroscopy",
            "molecules",
            "HITRAN",
            "hitemp",
            "exomol",
            "line-by-line",
        ],
        packages=find_packages(),
        install_requires=[
            "hitran-api",
            "beautifulsoup4",  # parse ExoMol website
            "lxml",  # parser used for ExoMol website
            "pyarrow",  # for the feather format (temporarily needed for ExoMol)
            "numpy",
            "scipy>=1.4.0",
            "matplotlib",  # ">=3.4.0" to suppress the Ruler warning, but only available for Python >= 3.7
            "seaborn",  # other matplotlib themes
            "cython",
            "pandas>=1.0.5",
            "plotly>=2.5.1",
            "numba",
            "mpldatacursor",
            "astropy",  # Unit aware calculations
            "publib>=0.3.2",  # Plotting styles for Matplotlib
            "plotly>=2.5.1",  # for line survey HTML output
            "termcolor",  # terminal colors
            "configparser",
            "astroquery>=0.3.9",  # to fetch HITRAN databases
            "json-tricks>=3.15.0",  # to deal with non jsonable formats
            "tables",  # for pandas to HDF5 export
            "pytest",  # to run test suite
            "joblib",  # for parallel loading of SpecDatabase
            "numba",  # just-in-time compiler
            "psutil",  # for getting user RAM
            "tuna",  # to generate visual/interactive performance profiles
            "vaex",  # HDF5,
            "h5py",  # HDF5
            "habanero",  # CrossRef API to retrieve data from doi
        ],
        extras_require={
            "dev": [
                "numpydoc",  # for Jedi (autocompletion) to recognize
                "black>=20.8b1",  # for code-linting in accordance to PEP8
                "isort",  # for sorting imports
                "pre-commit",  # to enforce Black before each commit
            ]
        },
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
            "Topic :: Scientific/Engineering",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Operating System :: OS Independent",
        ],
        **get_ext_modules(with_binary),
        include_package_data=True,  # add non .py data files in MANIFEST.in
        # package_data={'radis': ['radis/phys/units.txt']},
        zip_safe=False,  # impossible as long as we have external files read with __file__ syntax
        platforms="any",
    )


# %% Run Main install routine
try:
    run_setup(with_binary=True)
except BuildFailed:
    import traceback

    traceback.print_exc()
    show_message(
        "WARNING: RADIS C-extension could not be compiled, speedups",
        " are not enabled.",
        "Failure information, if any, is above.",
        "Retrying the build without the C extension now.",
    )
    run_setup(with_binary=False)
    show_message(
        "WARNING: RADIS C-extension could not be compiled, speedups"
        " are not enabled.",
        "Plain-Python build succeeded.",
    )



# -*- coding: utf-8 -*-
"""Functions specific to the calculation of CO2 energy levels.

Notes
-----

@dev: you can define Python functions here, with hardcoded spectroscopic coefficients.
Or, you can use the spectroscopic coefficients given in ``radis/db/CO2`` under
JSON format.

In both cases, you will need to overwrite the :meth:`~radis.db.classes.ElectronicState.Erovib`
method. See ``radis.db.molecules.py``


----------------------------------------------------------------------------------------------
"""


from radis.levels.vibrating_rotor import (
    EvJ_uncoupled_vibrating_rotor,
    EvJah_uncoupled_vibrating_rotor,
)


def EvJ_co2(v1, v2, l2, v3, J, coeff_dict, remove_ZPE=True):
    """Rovibrational energies of CO2 calculated with
    :py:func:`~radis.levels.vibrating_rotor.EvJ_uncoupled_vibrating_rotor`

    Examples
    --------

    Calculate energy of CO2 first asymetric mode (v3 = 1) ::

        from radis.db.utils import get_herzberg_coefficients
        from radis.levels.energies_co2 import EvJ_co2
        herzberg = get_herzberg_coefficients('CO2', 1, 'X1SIGu+')

        # Now calculate energy
        E = EvJ_co2(0,0,0,1,0,herzberg)


    See Also
    --------

    :py:func:`~radis.levels.vibrating_rotors.EvJ_uncoupled_vibrating_rotor`

    """
    return EvJ_uncoupled_vibrating_rotor(
        v1,
        v2,
        l2,
        v3,
        J,
        coeff_dict=coeff_dict,
        gv1=1,
        gv2=2,
        gv3=1,
        remove_ZPE=remove_ZPE,
    )


def EvJah_co2(v1, v2, l2, v3, J, coeff_dict, remove_ZPE=True):
    """Harmonic and anharmonic parts of rovibrational energies of CO2
    calculated with
    :py:func:`~radis.levels.vibrating_rotor.EvJah_uncoupled_vibrating_rotor`

    See Also
    --------

    :py:func:`~radis.levels.vibrating_rotor.EvJah_uncoupled_vibrating_rotor`
    """
    return EvJah_uncoupled_vibrating_rotor(
        v1, v2, l2, v3, J, coeff_dict=coeff_dict, remove_ZPE=remove_ZPE
    )
